# -*- encoding: utf-8 -*-
"""This files exposes non-gevent Python original functions."""
import threading

import attr
import six

from ddtrace.internal import compat
from ddtrace.internal import forksafe


try:
    import gevent.monkey
except ImportError:

    def get_original(module, func):
        return getattr(__import__(module), func)

    def is_module_patched(module):
        return False


else:
    get_original = gevent.monkey.get_original
    is_module_patched = gevent.monkey.is_module_patched


sleep = get_original("time", "sleep")

try:
    # Python ≥ 3.8
    threading_get_native_id = get_original("threading", "get_native_id")
except AttributeError:
    threading_get_native_id = None

start_new_thread = get_original(six.moves._thread.__name__, "start_new_thread")
thread_get_ident = get_original(six.moves._thread.__name__, "get_ident")
Thread = get_original("threading", "Thread")
Lock = get_original("threading", "Lock")

if six.PY2 and is_module_patched("threading"):
    _allocate_lock = get_original("threading", "_allocate_lock")
    _threading_RLock = get_original("threading", "_RLock")
    _threading_Verbose = get_original("threading", "_Verbose")

    class _RLock(_threading_RLock):
        """Patched RLock to ensure threading._allocate_lock is called rather than
        gevent.threading._allocate_lock if patching has occurred. This is not
        necessary in Python 3 where the RLock function uses the _CRLock so is
        unaffected by gevent patching.
        """

        def __init__(self, verbose=None):
            # We want to avoid calling the RLock init as it will allocate a gevent lock
            # That means we have to reproduce the code from threading._RLock.__init__ here
            # https://github.com/python/cpython/blob/8d21aa21f2cbc6d50aab3f420bb23be1d081dac4/Lib/threading.py#L132-L136
            _threading_Verbose.__init__(self, verbose)
            self.__block = _allocate_lock()
            self.__owner = None
            self.__count = 0

    def RLock(*args, **kwargs):
        return _RLock(*args, **kwargs)


else:
    # We do not patch RLock in Python 3 however for < 3.7 the C implementation of
    # RLock might not be available as the _thread module is optional.  In that
    # case, the Python implementation will be used. This means there is still
    # the possibility that RLock in Python 3 will cause problems for gevent with
    # ddtrace profiling enabled though it remains an open question when that
    # would be the case for the supported platforms.
    # https://github.com/python/cpython/blob/c19983125a42a4b4958b11a26ab5e03752c956fc/Lib/threading.py#L38-L41
    # https://github.com/python/cpython/blob/c19983125a42a4b4958b11a26ab5e03752c956fc/Doc/library/_thread.rst#L26-L27
    RLock = get_original("threading", "RLock")


is_threading_patched = is_module_patched("threading")

if is_threading_patched:

    @attr.s
    class DoubleLock(object):
        """A lock that prevent concurrency from a gevent coroutine and from a threading.Thread at the same time."""

        # This is a gevent-patched threading.Lock (= a gevent Lock)
        _lock = attr.ib(factory=forksafe.Lock, init=False, repr=False)
        # This is a unpatched threading.Lock (= a real threading.Lock)
        _thread_lock = attr.ib(factory=lambda: forksafe.ResetObject(Lock), init=False, repr=False)

        def acquire(self):
            # type: () -> None
            # You cannot acquire a gevent-lock from another thread if it has been acquired already:
            # make sure we exclude the gevent-lock from being acquire by another thread by using a thread-lock first.
            self._thread_lock.acquire()
            self._lock.acquire()

        def release(self):
            # type: () -> None
            self._lock.release()
            self._thread_lock.release()

        def __enter__(self):
            # type: () -> DoubleLock
            self.acquire()
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.release()


else:
    DoubleLock = threading.Lock  # type:  ignore[misc,assignment]


if is_threading_patched:
    # NOTE: bold assumption: this module is always imported by the MainThread.
    # The python `threading` module makes that assumption and it's beautiful we're going to do the same.
    # We don't have the choice has we can't access the original MainThread
    main_thread_id = thread_get_ident()
else:
    main_thread_id = compat.main_thread.ident



from functools import partial
import sys
import typing
import unittest

import mock
import pytest

from ddtrace.internal.utils import ArgumentError
from ddtrace.internal.utils import get_argument_value
from ddtrace.internal.utils import set_argument_value
from ddtrace.internal.utils import time
from ddtrace.internal.utils.cache import cached
from ddtrace.internal.utils.cache import cachedmethod
from ddtrace.internal.utils.formats import asbool
from ddtrace.internal.utils.formats import parse_tags_str
from ddtrace.internal.utils.importlib import func_name
from ddtrace.internal.utils.version import parse_version


class TestUtils(unittest.TestCase):
    def test_asbool(self):
        # ensure the value is properly cast
        self.assertTrue(asbool("True"))
        self.assertTrue(asbool("true"))
        self.assertTrue(asbool("1"))
        self.assertFalse(asbool("False"))
        self.assertFalse(asbool("false"))
        self.assertFalse(asbool(None))
        self.assertFalse(asbool(""))
        self.assertTrue(asbool(True))
        self.assertFalse(asbool(False))


_LOG_ERROR_MALFORMED_TAG = "Malformed tag in tag pair '%s' from tag string '%s'."
_LOG_ERROR_MALFORMED_TAG_STRING = "Malformed tag string with tags not separated by comma or space '%s'."
_LOG_ERROR_FAIL_SEPARATOR = (
    "Failed to find separator for tag string: '%s'.\n"
    "Tag strings must be comma or space separated:\n"
    "  key1:value1,key2:value2\n"
    "  key1:value1 key2:value2"
)


@pytest.mark.parametrize(
    "tag_str,expected_tags,error_calls",
    [
        ("", dict(), None),
        ("key:val", dict(key="val"), None),
        ("key:val,key2:val2", dict(key="val", key2="val2"), None),
        ("key:val,key2:val2,key3:1234.23", dict(key="val", key2="val2", key3="1234.23"), None),
        ("key:val key2:val2 key3:1234.23", dict(key="val", key2="val2", key3="1234.23"), None),
        ("key: val", dict(key=" val"), None),
        ("key key: val", {"key key": " val"}, None),
        ("key: val,key2:val2", dict(key=" val", key2="val2"), None),
        (" key: val,key2:val2", {"key": " val", "key2": "val2"}, None),
        ("key key2:val1", {"key key2": "val1"}, None),
        ("key:val key2:val:2", {"key": "val", "key2": "val:2"}, None),
        (
            "key:val,key2:val2 key3:1234.23",
            dict(),
            [mock.call(_LOG_ERROR_FAIL_SEPARATOR, "key:val,key2:val2 key3:1234.23")],
        ),
        ("key:val key2:val2 key3: ", dict(key="val", key2="val2", key3=""), None),
        (
            "key:val key2:val 2",
            dict(key="val", key2="val"),
            [mock.call(_LOG_ERROR_MALFORMED_TAG, "2", "key:val key2:val 2")],
        ),
        (
            "key: val key2:val2 key3:val3",
            {"key": "", "key2": "val2", "key3": "val3"},
            [mock.call(_LOG_ERROR_MALFORMED_TAG, "val", "key: val key2:val2 key3:val3")],
        ),
        ("key:,key3:val1,", dict(key3="val1", key=""), None),
        (",", dict(), [mock.call(_LOG_ERROR_FAIL_SEPARATOR, "")]),
        (":,:", dict(), [mock.call(_LOG_ERROR_FAIL_SEPARATOR, ":,:")]),
        ("key,key2:val1", {"key2": "val1"}, [mock.call(_LOG_ERROR_MALFORMED_TAG, "key", "key,key2:val1")]),
        ("key2:val1:", {"key2": "val1:"}, None),
        ("key,key2,key3", dict(), [mock.call(_LOG_ERROR_FAIL_SEPARATOR, "key,key2,key3")]),
        ("foo:bar,foo:baz", dict(foo="baz"), None),
        ("hash:asd url:https://github.com/foo/bar", dict(hash="asd", url="https://github.com/foo/bar"), None),
    ],
)
def test_parse_env_tags(tag_str, expected_tags, error_calls):
    with mock.patch("ddtrace.internal.utils.formats.log") as log:
        tags = parse_tags_str(tag_str)
        assert tags == expected_tags
        if error_calls:
            assert log.error.call_count == len(error_calls), log.error.call_args_list
            log.error.assert_has_calls(error_calls)
        else:
            assert log.error.call_count == 0, log.error.call_args_list


def test_no_states():
    watch = time.StopWatch()
    with pytest.raises(RuntimeError):
        watch.stop()


def test_start_stop():
    watch = time.StopWatch()
    watch.start()
    watch.stop()


def test_start_stop_elapsed():
    watch = time.StopWatch()
    watch.start()
    watch.stop()
    e = watch.elapsed()
    assert e > 0
    watch.start()
    assert watch.elapsed() != e


def test_no_elapsed():
    watch = time.StopWatch()
    with pytest.raises(RuntimeError):
        watch.elapsed()


def test_elapsed():
    watch = time.StopWatch()
    watch.start()
    watch.stop()
    assert watch.elapsed() > 0


def test_context_manager():
    with time.StopWatch() as watch:
        pass
    assert watch.elapsed() > 0


class SomethingCallable(object):
    """
    A dummy class that implements __call__().
    """

    value = 42

    def __call__(self):
        return "something"

    def me(self):
        return self

    @staticmethod
    def add(a, b):
        return a + b

    @classmethod
    def answer(cls):
        return cls.value


def some_function():
    """
    A function doing nothing.
    """
    return "nothing"


def minus(a, b):
    return a - b


minus_two = partial(minus, b=2)  # partial funcs need special handling (no module)

# disabling flake8 test below, yes, declaring a func like this is bad, we know
plus_three = lambda x: x + 3  # noqa: E731


class TestContrib(object):
    """
    Ensure that contrib utility functions handles corner cases
    """

    def test_func_name(self):
        # check that func_name works on anything callable, not only funcs.
        assert "nothing" == some_function()
        assert "tests.tracer.test_utils.some_function" == func_name(some_function)

        f = SomethingCallable()
        assert "something" == f()
        assert "tests.tracer.test_utils.SomethingCallable" == func_name(f)

        assert f == f.me()
        assert "tests.tracer.test_utils.me" == func_name(f.me)
        assert 3 == f.add(1, 2)
        assert "tests.tracer.test_utils.add" == func_name(f.add)
        assert 42 == f.answer()
        assert "tests.tracer.test_utils.answer" == func_name(f.answer)

        assert "tests.tracer.test_utils.minus" == func_name(minus)
        assert 5 == minus_two(7)
        if sys.version_info >= (3, 10, 0):
            assert "functools.partial" == func_name(minus_two)
        else:
            assert "partial" == func_name(minus_two)
        assert 10 == plus_three(7)
        assert "tests.tracer.test_utils.<lambda>" == func_name(plus_three)


@pytest.mark.parametrize(
    "args,kwargs,pos,kw,expected",
    [
        ([], {"foo": 42, "bar": "snafu"}, 0, "foo", 42),
        ([], {"foo": 42, "bar": "snafu"}, 1, "bar", "snafu"),
        ([42], {"bar": "snafu"}, 0, "foo", 42),
        ([42, "snafu"], {}, 1, "bar", "snafu"),
    ],
)
def test_infer_arg_value_hit(args, kwargs, pos, kw, expected):
    assert get_argument_value(args, kwargs, pos, kw) == expected


@pytest.mark.parametrize(
    "args,kwargs,pos,kw",
    [
        ([], {}, 0, "foo"),
        ([], {}, 1, "bar"),
    ],
)
def test_infer_arg_value_miss(args, kwargs, pos, kw):
    with pytest.raises(ArgumentError) as e:
        get_argument_value(args, kwargs, pos, kw)
        assert e.value == "%s (at position %d)" % (kw, pos)


@pytest.mark.parametrize(
    "args,kwargs,pos,kw,new_value",
    [
        ((), {"foo": 42, "bar": "snafu"}, 0, "foo", 4442),
        ((), {"foo": 42, "bar": "snafu"}, 1, "bar", "new_snafu"),
        ((42,), {"bar": "snafu"}, 0, "foo", 442),
        ((42, "snafu"), {}, 1, "bar", "snafu_new"),
    ],
)
def test_set_argument_value(args, kwargs, pos, kw, new_value):
    new_args, new_kwargs = set_argument_value(args, kwargs, pos, kw, new_value)

    if kw in kwargs:
        assert new_kwargs[kw] == new_value
    else:
        assert new_args[pos] == new_value


@pytest.mark.parametrize(
    "args,kwargs,pos,kw,value",
    [
        ([], {}, 0, "foo", "val"),
        ([], {}, 1, "bar", "val"),
    ],
)
def test_set_invalid_argument_value(args, kwargs, pos, kw, value):
    with pytest.raises(ArgumentError) as e:
        set_argument_value(args, kwargs, pos, kw, value)
        assert e.value == "%s (at position %d) is invalid" % (kw, pos)


def cached_test_recipe(expensive, cheap, witness, cache_size):
    assert cheap("Foo") == expensive("Foo")
    assert cheap("Foo") == expensive("Foo")

    witness.assert_called_with("Foo")
    assert witness.call_count == 1

    cheap.invalidate()

    for i in range(cache_size >> 1):
        cheap("Foo%d" % i)

    assert witness.call_count == 1 + (cache_size >> 1)

    for i in range(cache_size):
        cheap("Foo%d" % i)

    assert witness.call_count == 1 + cache_size

    MAX_FOO = "Foo%d" % (cache_size - 1)

    cheap("last drop")  # Forces least frequent elements out of the cache
    assert witness.call_count == 2 + cache_size

    cheap(MAX_FOO)  # Check MAX_FOO was dropped
    assert witness.call_count == 3 + cache_size

    cheap("last drop")  # Check last drop was retained
    assert witness.call_count == 3 + cache_size


def test_cached():
    witness = mock.Mock()
    cache_size = 128

    def expensive(key):
        return key[::-1].lower()

    @cached(cache_size)
    def cheap(key):
        witness(key)
        return expensive(key)

    cached_test_recipe(expensive, cheap, witness, cache_size)


def test_cachedmethod():
    witness = mock.Mock()
    cache_size = 128

    def expensive(key):
        return key[::-1].lower()

    class Foo(object):
        @cachedmethod(cache_size)
        def cheap(self, key):
            witness(key)
            return expensive(key)

    cached_test_recipe(expensive, Foo().cheap, witness, cache_size)


@pytest.mark.parametrize(
    "version_str,expected",
    [
        ("5", (5, 0, 0)),
        ("0.5", (0, 5, 0)),
        ("0.5.0", (0, 5, 0)),
        ("1.0.0", (1, 0, 0)),
        ("1.2.0", (1, 2, 0)),
        ("1.2.8", (1, 2, 8)),
        ("2.0.0rc1", (2, 0, 0)),
        ("2.0.0-rc1", (2, 0, 0)),
        ("2.0.0 here be dragons", (2, 0, 0)),
        ("2020.6.19", (2020, 6, 19)),
        ("beta 1.0.0", (0, 0, 0)),
        ("no version found", (0, 0, 0)),
        ("", (0, 0, 0)),
    ],
)
def test_parse_version(version_str, expected):
    # type: (str, typing.Tuple[int, int, int]) -> None
    """Ensure parse_version helper properly parses versions"""
    assert parse_version(version_str) == expected



import abc

import six


class AbsStuff(six.with_metaclass(abc.ABCMeta)):
    @abc.abstractmethod
    def foo(self):
        pass

    @abc.abstractmethod
    def bar(self):
        pass


class ConcrStuff(AbsStuff):
    def foo(self):
        return "foo"

    def bar(self):
        return "bar"



# coding: utf-8
from collections import defaultdict
import os
import typing

from ddsketch import LogCollapsingLowestDenseDDSketch
from ddsketch.pb.proto import DDSketchProto
import six
import tenacity

import ddtrace
from ddtrace import config
from ddtrace.span import _is_top_level

from . import SpanProcessor
from ...constants import SPAN_MEASURED_KEY
from .._encoding import packb
from ..agent import get_connection
from ..compat import get_connection_response
from ..compat import httplib
from ..forksafe import Lock
from ..hostname import get_hostname
from ..logger import get_logger
from ..periodic import PeriodicService
from ..writer import _human_size


if typing.TYPE_CHECKING:
    from typing import DefaultDict
    from typing import Dict
    from typing import List
    from typing import Optional
    from typing import Union

    from ddtrace import Span


log = get_logger(__name__)


def _is_measured(span):
    # type: (Span) -> bool
    """Return whether the span is flagged to be measured or not."""
    return span._metrics.get(SPAN_MEASURED_KEY) == 1


"""
To aggregate metrics for spans they need to be "uniquely" identified (as
best as possible). This enables the compression of stat points.

Aggregation can be done using primary and secondary attributes from the span
stored in a tuple which is hashable in Python.
"""
SpanAggrKey = typing.Tuple[
    str,  # name
    str,  # service
    str,  # resource
    str,  # type
    int,  # http status code
    bool,  # synthetics request
]


class SpanAggrStats(object):
    """Aggregated span statistics."""

    __slots__ = ("hits", "top_level_hits", "errors", "duration", "ok_distribution", "err_distribution")

    def __init__(self):
        self.hits = 0
        self.top_level_hits = 0
        self.errors = 0
        self.duration = 0
        # Match the relative accuracy of the sketch implementation used in the backend
        # which is 0.775%.
        self.ok_distribution = LogCollapsingLowestDenseDDSketch(0.00775, bin_limit=2048)
        self.err_distribution = LogCollapsingLowestDenseDDSketch(0.00775, bin_limit=2048)


def _span_aggr_key(span):
    # type: (Span) -> SpanAggrKey
    """Return a hashable key that can be used to aggregate similar spans."""
    service = span.service or ""
    resource = span.resource or ""
    _type = span.span_type or ""
    status_code = span.get_tag("http.status_code") or 0
    synthetics = span.context.dd_origin == "synthetics"
    return span.name, service, resource, _type, int(status_code), synthetics


class SpanStatsProcessorV06(PeriodicService, SpanProcessor):
    """SpanProcessor for computing, collecting and submitting span metrics to the Datadog Agent."""

    def __init__(self, agent_url, interval=None, timeout=1.0, retry_attempts=3):
        # type: (str, Optional[float], float, int) -> None
        if interval is None:
            interval = float(os.getenv("_DD_TRACE_STATS_WRITER_INTERVAL") or 10.0)
        super(SpanStatsProcessorV06, self).__init__(interval=interval)
        self._agent_url = agent_url
        self._endpoint = "/v0.6/stats"
        self._agent_endpoint = "%s%s" % (self._agent_url, self._endpoint)
        self._timeout = timeout
        # Have the bucket size match the interval in which flushes occur.
        self._bucket_size_ns = int(interval * 1e9)  # type: int
        self._buckets = defaultdict(
            lambda: defaultdict(SpanAggrStats)
        )  # type: DefaultDict[int, DefaultDict[SpanAggrKey, SpanAggrStats]]
        self._headers = {
            "Datadog-Meta-Lang": "python",
            "Datadog-Meta-Tracer-Version": ddtrace.__version__,
            "Content-Type": "application/msgpack",
        }  # type: Dict[str, str]
        self._hostname = six.ensure_text(get_hostname())
        self._lock = Lock()
        self._enabled = True
        self._retry_request = tenacity.Retrying(
            # Use a Fibonacci policy with jitter, same as AgentWriter.
            wait=tenacity.wait_random_exponential(
                multiplier=0.618 * self.interval / (1.618 ** retry_attempts) / 2, exp_base=1.618
            ),
            stop=tenacity.stop_after_attempt(retry_attempts),
            retry=tenacity.retry_if_exception_type((httplib.HTTPException, OSError, IOError)),
        )
        self.start()

    def on_span_start(self, span):
        # type: (Span) -> None
        pass

    def on_span_finish(self, span):
        # type: (Span) -> None
        if not self._enabled:
            return

        is_top_level = _is_top_level(span)
        if not is_top_level and not _is_measured(span):
            return

        with self._lock:
            # Align the span into the corresponding stats bucket
            assert span.duration_ns is not None
            span_end_ns = span.start_ns + span.duration_ns
            bucket_time_ns = span_end_ns - (span_end_ns % self._bucket_size_ns)
            aggr_key = _span_aggr_key(span)
            stats = self._buckets[bucket_time_ns][aggr_key]

            stats.hits += 1
            stats.duration += span.duration_ns
            if is_top_level:
                stats.top_level_hits += 1
            if span.error:
                stats.errors += 1
                stats.err_distribution.add(span.duration_ns)
            else:
                stats.ok_distribution.add(span.duration_ns)

    def _serialize_buckets(self):
        # type: () -> List[Dict]
        """Serialize and update the buckets.

        The current bucket is left in case any other spans are added.
        """
        serialized_buckets = []
        serialized_bucket_keys = []
        for bucket_time_ns, bucket in self._buckets.items():
            bucket_aggr_stats = []
            serialized_bucket_keys.append(bucket_time_ns)

            for aggr_key, stat_aggr in bucket.items():
                name, service, resource, _type, http_status, synthetics = aggr_key
                serialized_bucket = {
                    u"Name": six.ensure_text(name),
                    u"Resource": six.ensure_text(resource),
                    u"Synthetics": synthetics,
                    u"HTTPStatusCode": http_status,
                    u"Hits": stat_aggr.hits,
                    u"TopLevelHits": stat_aggr.top_level_hits,
                    u"Duration": stat_aggr.duration,
                    u"Errors": stat_aggr.errors,
                    u"OkSummary": DDSketchProto.to_proto(stat_aggr.ok_distribution).SerializeToString(),
                    u"ErrorSummary": DDSketchProto.to_proto(stat_aggr.err_distribution).SerializeToString(),
                }
                if service:
                    serialized_bucket[u"Service"] = six.ensure_text(service)
                if _type:
                    serialized_bucket[u"Type"] = six.ensure_text(_type)
                bucket_aggr_stats.append(serialized_bucket)
            serialized_buckets.append(
                {
                    u"Start": bucket_time_ns,
                    u"Duration": self._bucket_size_ns,
                    u"Stats": bucket_aggr_stats,
                }
            )

        # Clear out buckets that have been serialized
        for key in serialized_bucket_keys:
            del self._buckets[key]

        return serialized_buckets

    def _flush_stats(self, payload):
        # type: (bytes) -> None
        try:
            conn = get_connection(self._agent_url, self._timeout)
            conn.request("PUT", self._endpoint, payload, self._headers)
            resp = get_connection_response(conn)
        except Exception:
            log.error("failed to submit span stats to the Datadog agent at %s", self._agent_endpoint, exc_info=True)
            raise
        else:
            if resp.status == 404:
                log.error(
                    "Datadog agent does not support tracer stats computation, disabling, please upgrade your agent"
                )
                self._enabled = False
                return
            elif resp.status >= 400:
                log.error(
                    "failed to send stats payload, %s (%s) (%s) response from Datadog agent at %s",
                    resp.status,
                    resp.reason,
                    resp.read(),
                    self._agent_endpoint,
                )
            else:
                log.info("sent %s to %s", _human_size(len(payload)), self._agent_endpoint)

    def periodic(self):
        # type: (...) -> None

        with self._lock:
            serialized_stats = self._serialize_buckets()

        if not serialized_stats:
            # No stats to report, short-circuit.
            return
        raw_payload = {
            u"Stats": serialized_stats,
            u"Hostname": self._hostname,
        }  # type: Dict[str, Union[List[Dict], str]]
        if config.env:
            raw_payload[u"Env"] = six.ensure_text(config.env)
        if config.version:
            raw_payload[u"Version"] = six.ensure_text(config.version)

        payload = packb(raw_payload)
        try:
            self._retry_request(self._flush_stats, payload)
        except tenacity.RetryError:
            log.error("retry limit exceeded submitting span stats to the Datadog agent at %s", self._agent_endpoint)

    def shutdown(self, timeout):
        # type: (Optional[float]) -> None
        self.periodic()
        self.stop(timeout)



from ddtrace.constants import _SINGLE_SPAN_SAMPLING_MAX_PER_SEC
from ddtrace.constants import _SINGLE_SPAN_SAMPLING_MECHANISM
from ddtrace.constants import _SINGLE_SPAN_SAMPLING_RATE
from ddtrace.internal.sampling import SamplingMechanism
from ddtrace.internal.sampling import SpanSamplingRule

from ..utils import DummyTracer


def assert_sampling_decision_tags(span, sample_rate=1.0, mechanism=SamplingMechanism.SPAN_SAMPLING_RULE, limit=None):
    assert span.get_metric(_SINGLE_SPAN_SAMPLING_RATE) == sample_rate
    assert span.get_metric(_SINGLE_SPAN_SAMPLING_MECHANISM) == mechanism
    assert span.get_metric(_SINGLE_SPAN_SAMPLING_MAX_PER_SEC) == limit


def traced_function(rule, name="test_name", service="test_service"):
    tracer = DummyTracer()
    with tracer.trace(name) as span:
        span.service = service
        if rule.match(span):
            rule.sample(span)
    return span


def test_single_span_rule_no_match_empty_strings():
    rule = SpanSamplingRule(service="", name="", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule)
    assert_sampling_decision_tags(span, sample_rate=None, mechanism=None, limit=None)


def test_single_span_rule_no_match_service():
    rule = SpanSamplingRule(service="wrong_service_name", name="test_name", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule)

    assert_sampling_decision_tags(span, sample_rate=None, mechanism=None, limit=None)


def test_single_span_rule_no_match_name():
    rule = SpanSamplingRule(service="test_service", name="wrong_operation_name", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule)

    assert_sampling_decision_tags(span, sample_rate=None, mechanism=None, limit=None)


def test_single_span_rule_no_match_only_service():
    rule = SpanSamplingRule(service="wrong_service_name", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule)

    assert_sampling_decision_tags(span, sample_rate=None, mechanism=None, limit=None)


def test_single_span_rule_no_match_no_span_name_or_service():
    rule = SpanSamplingRule(service="test_service", name="test_name", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule, name=None, service=None)

    assert_sampling_decision_tags(span, sample_rate=None, mechanism=None, limit=None)


def test_single_span_rule_only_span_service():
    rule = SpanSamplingRule(service="test_service", name="test_name", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule, name=None)

    assert_sampling_decision_tags(span, sample_rate=None, mechanism=None, limit=None)


def test_single_span_rule_only_span_name():
    rule = SpanSamplingRule(service="test_service", name="test_name", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule, service=None)

    assert_sampling_decision_tags(span, sample_rate=None, mechanism=None, limit=None)


def test_single_span_rule_no_match_only_name():
    rule = SpanSamplingRule(name="wrong_operation_name", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule)
    assert_sampling_decision_tags(span, sample_rate=None, mechanism=None, limit=None)


def test_single_span_rule_match():
    rule = SpanSamplingRule(service="test_service", name="test_name", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule)

    assert_sampling_decision_tags(span)


def test_single_span_rule_match_only_service():
    rule = SpanSamplingRule(service="test_service", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule)

    assert_sampling_decision_tags(span)


def test_single_span_rule_match_only_name():
    rule = SpanSamplingRule(name="test_name", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule)

    assert_sampling_decision_tags(span)


# More extensive testing of the matching is done in test_glob_matcher.py
def test_single_span_rule_match_wildcards():
    rule = SpanSamplingRule(service="test_*", name="test?????", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule)

    assert_sampling_decision_tags(span)


def test_single_span_rule_no_match_wildcards():
    rule = SpanSamplingRule(service="*test_", name="test_nam??", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule)

    assert_sampling_decision_tags(span, sample_rate=None, mechanism=None, limit=None)


def test_single_span_rule_unsupported_pattern_bracket_expansion():
    rule = SpanSamplingRule(service="test_servic[a-z]+", name="test_name", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule)

    assert_sampling_decision_tags(span, sample_rate=None, mechanism=None, limit=None)


def test_single_span_rule_unsupported_pattern_escape_character():
    rule = SpanSamplingRule(service="test_servic[?]", name="test_name", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule)

    assert_sampling_decision_tags(span, sample_rate=None, mechanism=None, limit=None)


def test_single_span_rule_unsupported_pattern_bracket_expansion_literal_evaluation():
    rule = SpanSamplingRule(service="test_servic[a-z]+", name="test_name", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule, service="test_servic[a-z]+")

    assert_sampling_decision_tags(span)


def test_single_span_rule_unsupported_pattern_escape_character_literal_evaluation():
    rule = SpanSamplingRule(service="test_servic[?]", name="test_name", sample_rate=1.0, max_per_second=-1)
    span = traced_function(rule, service="test_servic[?]")

    assert_sampling_decision_tags(span)


def test_multiple_span_rule_match():
    rule = SpanSamplingRule(service="test_service", name="test_name", sample_rate=1.0, max_per_second=-1)
    for i in range(10):
        span = traced_function(rule)

        assert_sampling_decision_tags(span)


def test_max_per_sec_0():
    rule = SpanSamplingRule(service="test_service", name="test_name", sample_rate=1.0, max_per_second=0)
    for i in range(10):
        span = traced_function(rule)
        # None of the single span sampling tags should be added to the span
        # if it's dropped by the single span rate limiter
        assert_sampling_decision_tags(span, sample_rate=None, mechanism=None, limit=None)


def test_match_max_per_sec():
    rule = SpanSamplingRule(service="test_service", name="test_name", sample_rate=1.0, max_per_second=2)
    # Make spans till we hit the limit, then make a span while the limit is hit and make sure tags were not added.
    while True:
        span = traced_function(rule)
        if rule._limiter._is_allowed(span.start_ns):
            assert_sampling_decision_tags(span, limit=2)
        else:
            break

    rate_limited_span = traced_function(rule)

    assert_sampling_decision_tags(rate_limited_span, sample_rate=None, mechanism=None, limit=None)



from setuptools import setup, find_packages

setup(
    name="pserve_test_app",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        "paste.app_factory": [
            "main = app:main",
        ],
    },
)



import asyncio
import collections
import os
import sys

import pytest

from ddtrace.profiling import _asyncio
from ddtrace.profiling import profiler
from ddtrace.profiling.collector import stack_event

from . import _asyncio_compat


TESTING_GEVENT = os.getenv("DD_PROFILE_TEST_GEVENT", False)


@pytest.mark.skipif(not _asyncio_compat.PY36_AND_LATER, reason="Python > 3.5 needed")
def test_asyncio(tmp_path, monkeypatch) -> None:
    sleep_time = 0.2
    sleep_times = 5

    async def stuff() -> None:
        await asyncio.sleep(sleep_time)

    async def hello() -> None:
        t1 = _asyncio_compat.create_task(stuff(), name="sleep 1")
        t2 = _asyncio_compat.create_task(stuff(), name="sleep 2")
        for _ in range(sleep_times):
            await stuff()
        return (t1, t2)

    monkeypatch.setenv("DD_PROFILING_CAPTURE_PCT", "100")
    monkeypatch.setenv("DD_PROFILING_OUTPUT_PPROF", str(tmp_path / "pprof"))
    # start a complete profiler so asyncio policy is setup
    p = profiler.Profiler()
    p.start()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    if _asyncio_compat.PY38_AND_LATER:
        maintask = loop.create_task(hello(), name="main")
    else:
        maintask = loop.create_task(hello())
    t1, t2 = loop.run_until_complete(maintask)
    events = p._profiler._recorder.reset()
    p.stop()

    wall_time_ns = collections.defaultdict(lambda: 0)

    t1_name = _asyncio._task_get_name(t1)
    t2_name = _asyncio._task_get_name(t2)

    cpu_time_found = False
    stack_sample_events = events[stack_event.StackSampleEvent]
    for event in stack_sample_events:

        wall_time_ns[event.task_name] += event.wall_time_ns

        # This assertion does not work reliably on Python < 3.7
        if _asyncio_compat.PY37_AND_LATER:
            if event.task_name == "main":
                assert event.thread_name == "MainThread"
                assert event.frames == [(__file__, 30, "hello", "")]
                assert event.nframes == 1
            elif event.task_name == t1_name:
                assert event.thread_name == "MainThread"
                assert event.frames == [(__file__, 24, "stuff", "")]
                assert event.nframes == 1
            elif event.task_name == t2_name:
                assert event.thread_name == "MainThread"
                assert event.frames == [(__file__, 24, "stuff", "")]
                assert event.nframes == 1

        if event.thread_name == "MainThread" and (
            # The task name is empty in asyncio (it's not a task) but the main thread is seen as a task in gevent
            (event.task_name is None and not TESTING_GEVENT)
            or (event.task_name == "MainThread" and TESTING_GEVENT)
        ):
            # Make sure we account CPU time
            if event.cpu_time_ns > 0:
                cpu_time_found = True

            for frame in event.frames:
                if frame[0] == __file__ and frame[2] == "test_asyncio":
                    break
            else:
                pytest.fail("unable to find expected main thread frame: %r" % event.frames)

    if _asyncio_compat.PY38_AND_LATER:
        # We don't know the name of this task for Python < 3.8
        assert wall_time_ns["main"] > 0, (wall_time_ns, stack_sample_events)

    assert wall_time_ns[t1_name] > 0
    assert wall_time_ns[t2_name] > 0
    if sys.platform != "win32":
        # Windows seems to get 0 CPU for this
        assert cpu_time_found



from __future__ import division

import re
import unittest

import mock
import pytest

from ddtrace.constants import AUTO_KEEP
from ddtrace.constants import AUTO_REJECT
from ddtrace.constants import SAMPLE_RATE_METRIC_KEY
from ddtrace.constants import SAMPLING_AGENT_DECISION
from ddtrace.constants import SAMPLING_LIMIT_DECISION
from ddtrace.constants import SAMPLING_PRIORITY_KEY
from ddtrace.constants import SAMPLING_RULE_DECISION
from ddtrace.constants import USER_KEEP
from ddtrace.constants import USER_REJECT
from ddtrace.internal.compat import iteritems
from ddtrace.internal.rate_limiter import RateLimiter
from ddtrace.internal.sampling import SAMPLING_DECISION_TRACE_TAG_KEY
from ddtrace.internal.sampling import SamplingMechanism
from ddtrace.sampler import AllSampler
from ddtrace.sampler import DatadogSampler
from ddtrace.sampler import RateByServiceSampler
from ddtrace.sampler import RateSampler
from ddtrace.sampler import SamplingRule
from ddtrace.span import Span

from ..utils import DummyTracer
from ..utils import override_env


@pytest.fixture
def dummy_tracer():
    return DummyTracer()


def assert_sampling_decision_tags(
    span,
    agent=None,
    limit=None,
    rule=None,
    sampling_priority=None,
    trace_tag=None,
):
    """Check span attribute given an expected sampling decision

    :param agent: expected agent rate ``_dd.agent_psr``
    :param limit: expected rate limit ``_dd.limit_psr``
    :param rule: expected sampler rule rate ``_dd.rule_psr``
    :param sampling_priority: expected sampling priority ``_sampling_priority_v1``
    :param trace_tag: expected sampling decision trace tag ``_dd.p.dm``. Format is ``-{SAMPLINGMECHANISM}``.
    """
    metric_agent = span.get_metric(SAMPLING_AGENT_DECISION)
    metric_limit = span.get_metric(SAMPLING_LIMIT_DECISION)
    metric_rule = span.get_metric(SAMPLING_RULE_DECISION)
    metric_sampling_priority = span.get_metric(SAMPLING_PRIORITY_KEY)
    if agent:
        assert metric_agent == agent
    if limit:
        assert metric_limit == limit
    if rule:
        assert metric_rule == rule
    if sampling_priority:
        assert metric_sampling_priority == sampling_priority

    if trace_tag:
        assert span.context._meta[SAMPLING_DECISION_TRACE_TAG_KEY] == trace_tag
    else:
        assert SAMPLING_DECISION_TRACE_TAG_KEY not in span.context._meta


def create_span(tracer=None, name="test.span", service=""):
    tracer = tracer or DummyTracer()
    span = tracer.trace(name=name, service=service)
    span.finish()
    return span


class RateSamplerTest(unittest.TestCase):
    def test_set_sample_rate(self):
        sampler = RateSampler()
        assert sampler.sample_rate == 1.0

        for rate in [0.001, 0.01, 0.1, 0.25, 0.5, 0.75, 0.99999999, 1.0, 1]:
            sampler.set_sample_rate(rate)
            assert sampler.sample_rate == float(rate)

            sampler.set_sample_rate(str(rate))
            assert sampler.sample_rate == float(rate)

    def test_set_sample_rate_str(self):
        sampler = RateSampler()
        sampler.set_sample_rate("0.5")
        assert sampler.sample_rate == 0.5

    def test_sample_rate_deviation(self):
        for sample_rate in [0.1, 0.25, 0.5, 1]:
            tracer = DummyTracer()

            tracer._sampler = RateSampler(sample_rate)

            iterations = int(1e4 / sample_rate)

            for i in range(iterations):
                span = tracer.trace(str(i))
                span.finish()

            samples = tracer.pop()

            # We must have at least 1 sample, check that it has its sample rate properly assigned
            assert samples[0].get_metric(SAMPLE_RATE_METRIC_KEY) == sample_rate

            # Less than 5% deviation when 'enough' iterations (arbitrary, just check if it converges)
            deviation = abs(len(samples) - (iterations * sample_rate)) / (iterations * sample_rate)
            assert deviation < 0.05, "Deviation too high %f with sample_rate %f" % (deviation, sample_rate)

    def test_deterministic_behavior(self):
        """Test that for a given trace ID, the result is always the same"""
        tracer = DummyTracer()

        tracer._sampler = RateSampler(0.5)

        for i in range(10):
            span = tracer.trace(str(i))
            span.finish()

            samples = tracer.pop()
            assert len(samples) <= 1, "there should be 0 or 1 spans"
            sampled = 1 == len(samples)
            for j in range(10):
                other_span = Span(str(i), trace_id=span.trace_id)
                assert sampled == tracer._sampler.sample(
                    other_span
                ), "sampling should give the same result for a given trace_id"

    def test_negative_sample_rate_raises_error(self):
        tracer = DummyTracer()
        with pytest.raises(ValueError, match="sample_rate of -0.5 is negative"):
            tracer._sampler = RateSampler(sample_rate=-0.5)

    def test_sample_rate_0_does_not_reset_to_1(self):
        # Regression test for case where a sample rate of 0 caused the sample rate to be reset to 1
        tracer = DummyTracer()
        tracer._sampler = RateSampler(sample_rate=0)
        assert tracer._sampler.sample_rate == 0


class RateByServiceSamplerTest(unittest.TestCase):
    def test_default_key(self):
        assert (
            "service:,env:" == RateByServiceSampler._default_key
        ), "default key should correspond to no service and no env"

    def test_key(self):
        assert RateByServiceSampler._default_key == RateByServiceSampler._key()
        assert "service:mcnulty,env:" == RateByServiceSampler._key(service="mcnulty")
        assert "service:,env:test" == RateByServiceSampler._key(env="test")
        assert "service:mcnulty,env:test" == RateByServiceSampler._key(service="mcnulty", env="test")
        assert "service:mcnulty,env:test" == RateByServiceSampler._key("mcnulty", "test")

    def test_sample_rate_deviation(self):
        for sample_rate in [0.1, 0.25, 0.5, 1]:
            tracer = DummyTracer()
            writer = tracer._writer
            tracer.configure(sampler=AllSampler())
            # We need to set the writer because tracer.configure overrides it,
            # indeed, as we enable priority sampling, we must ensure the writer
            # is priority sampling aware and pass it a reference on the
            # priority sampler to send the feedback it gets from the agent
            assert writer is not tracer._writer, "writer should have been updated by configure"
            tracer._priority_sampler.set_sample_rate(sample_rate)

            iterations = int(1e4 / sample_rate)

            for i in range(iterations):
                span = tracer.trace(str(i))
                span.finish()

            samples = tracer._writer.pop()
            samples_with_high_priority = 0
            for sample in samples:
                if sample.get_metric(SAMPLING_PRIORITY_KEY) is not None:
                    if sample.get_metric(SAMPLING_PRIORITY_KEY) > 0:
                        samples_with_high_priority += 1
                else:
                    assert 0 == sample.get_metric(
                        SAMPLING_PRIORITY_KEY
                    ), "when priority sampling is on, priority should be 0 when trace is to be dropped"
                assert_sampling_decision_tags(
                    sample,
                    agent=sample_rate,
                    # sampling decision made by agent
                    trace_tag="-1",
                )
            # We must have at least 1 sample, check that it has its sample rate properly assigned
            assert samples[0].get_metric(SAMPLE_RATE_METRIC_KEY) is None

            # Less than 5% deviation when 'enough' iterations (arbitrary, just check if it converges)
            deviation = abs(samples_with_high_priority - (iterations * sample_rate)) / (iterations * sample_rate)
            assert deviation < 0.05, "Deviation too high %f with sample_rate %f" % (deviation, sample_rate)

    def test_update_rate_by_service_sample_rates(self):
        cases = [
            {
                "service:,env:": 1,
            },
            {
                "service:,env:": 1,
                "service:mcnulty,env:dev": 0.33,
                "service:postgres,env:dev": 0.7,
            },
            {
                "service:,env:": 1,
                "service:mcnulty,env:dev": 0.25,
                "service:postgres,env:dev": 0.5,
                "service:redis,env:prod": 0.75,
            },
        ]

        tracer = DummyTracer()
        tracer.configure(sampler=AllSampler())
        priority_sampler = tracer._priority_sampler
        for case in cases:
            priority_sampler.update_rate_by_service_sample_rates(case)
            rates = {}
            for k, v in iteritems(priority_sampler._by_service_samplers):
                rates[k] = v.sample_rate
            assert case == rates, "%s != %s" % (case, rates)
        # It's important to also test in reverse mode for we want to make sure key deletion
        # works as well as key insertion (and doing this both ways ensures we trigger both cases)
        cases.reverse()
        for case in cases:
            priority_sampler.update_rate_by_service_sample_rates(case)
            rates = {}
            for k, v in iteritems(priority_sampler._by_service_samplers):
                rates[k] = v.sample_rate
            assert case == rates, "%s != %s" % (case, rates)


@pytest.mark.parametrize(
    "sample_rate,allowed",
    [
        # Min/max allowed values
        (0.0, True),
        (1.0, True),
        # Accepted boundaries
        (0.000001, True),
        (0.999999, True),
        # Outside the bounds
        (-0.000000001, False),
        (1.0000000001, False),
    ]
    + [
        # Try a bunch of decimal values between 0 and 1
        (1 / i, True)
        for i in range(1, 50)
    ]
    + [
        # Try a bunch of decimal values less than 0
        (-(1 / i), False)
        for i in range(1, 50)
    ]
    + [
        # Try a bunch of decimal values greater than 1
        (1 + (1 / i), False)
        for i in range(1, 50)
    ],
)
def test_sampling_rule_init_sample_rate(sample_rate, allowed):
    if allowed:
        rule = SamplingRule(sample_rate=sample_rate)
        assert rule.sample_rate == sample_rate
    else:
        with pytest.raises(ValueError):
            SamplingRule(sample_rate=sample_rate)


def test_sampling_rule_init_defaults():
    rule = SamplingRule(sample_rate=1.0)
    assert rule.sample_rate == 1.0
    assert rule.service == SamplingRule.NO_RULE
    assert rule.name == SamplingRule.NO_RULE


def test_sampling_rule_init():
    name_regex = re.compile(r"\.request$")

    rule = SamplingRule(
        sample_rate=0.0,
        # Value
        service="my-service",
        # Regex
        name=name_regex,
    )

    assert rule.sample_rate == 0.0
    assert rule.service == "my-service"
    assert rule.name == name_regex


@pytest.mark.parametrize(
    "rule_1,rule_2,expected",
    [
        # Sample rate only
        (SamplingRule(sample_rate=1.0), SamplingRule(sample_rate=1.0), True),
        (SamplingRule(sample_rate=0.5), SamplingRule(sample_rate=0.5), True),
        (SamplingRule(sample_rate=0.0), SamplingRule(sample_rate=0.0), True),
        (SamplingRule(sample_rate=0.5), SamplingRule(sample_rate=1.0), False),
        # Sample rate, and service name
        (SamplingRule(sample_rate=1.0, service="my-svc"), SamplingRule(sample_rate=1.0, service="my-svc"), True),
        (
            SamplingRule(sample_rate=1.0, service=re.compile("my-svc")),
            SamplingRule(sample_rate=1.0, service=re.compile("my-svc")),
            True,
        ),
        (SamplingRule(sample_rate=1.0, service="my-svc"), SamplingRule(sample_rate=1.0, service="other-svc"), False),
        (SamplingRule(sample_rate=1.0, service="my-svc"), SamplingRule(sample_rate=0.5, service="my-svc"), False),
        (
            SamplingRule(sample_rate=1.0, service=re.compile("my-svc")),
            SamplingRule(sample_rate=0.5, service=re.compile("my-svc")),
            False,
        ),
        (
            SamplingRule(sample_rate=1.0, service=re.compile("my-svc")),
            SamplingRule(sample_rate=1.0, service=re.compile("other")),
            False,
        ),
        # Sample rate, and operation name
        (
            SamplingRule(sample_rate=1.0, name="span.name"),
            SamplingRule(sample_rate=1.0, name="span.name"),
            True,
        ),
        (
            SamplingRule(sample_rate=1.0, name=re.compile("span.name")),
            SamplingRule(sample_rate=1.0, name=re.compile("span.name")),
            True,
        ),
        (
            SamplingRule(sample_rate=1.0, name=re.compile("span.name")),
            SamplingRule(sample_rate=1.0, name=re.compile("span.other")),
            False,
        ),
        (
            SamplingRule(sample_rate=1.0, name="span.name"),
            SamplingRule(sample_rate=0.5, name="span.name"),
            False,
        ),
        (SamplingRule(sample_rate=1.0, name="span.name"), SamplingRule(sample_rate=1.0, name="span.other"), False),
        (SamplingRule(sample_rate=1.0, name="span.name"), SamplingRule(sample_rate=0.5, name="span.name"), False),
        # Sample rate, service, and operation name
        (
            SamplingRule(sample_rate=1.0, service="my-svc", name="span.name"),
            SamplingRule(sample_rate=1.0, service="my-svc", name="span.name"),
            True,
        ),
        (
            SamplingRule(sample_rate=1.0, service="my-svc", name=re.compile("span.name")),
            SamplingRule(sample_rate=1.0, service="my-svc", name=re.compile("span.name")),
            True,
        ),
        (
            SamplingRule(sample_rate=1.0, service=re.compile("my-svc"), name=re.compile("span.name")),
            SamplingRule(sample_rate=1.0, service=re.compile("my-svc"), name=re.compile("span.name")),
            True,
        ),
        (
            SamplingRule(sample_rate=1.0, service="my-svc", name="span.name"),
            SamplingRule(sample_rate=0.5, service="my-svc", name="span.name"),
            False,
        ),
        (
            SamplingRule(sample_rate=1.0, service="my-svc", name="span.name"),
            SamplingRule(sample_rate=1.0, service="other", name="span.name"),
            False,
        ),
        (
            SamplingRule(sample_rate=1.0, service="my-svc", name="span.name"),
            SamplingRule(sample_rate=1.0, service="my-svc", name="span.other"),
            False,
        ),
    ],
)
def test_sampling_rule_eq(rule_1, rule_2, expected):
    result = rule_1 == rule_2
    assert result == expected


def test_sampling_rule_init_via_env():
    # Testing single sampling rule
    with override_env(dict(DD_TRACE_SAMPLING_RULES='[{"sample_rate":1.0,"service":"xyz","name":"abc"}]')):
        sampling_rule = DatadogSampler().rules
        assert sampling_rule[0].sample_rate == 1.0
        assert sampling_rule[0].service == "xyz"
        assert sampling_rule[0].name == "abc"
        assert len(sampling_rule) == 1

    # Testing multiple sampling rules
    with override_env(
        dict(
            DD_TRACE_SAMPLING_RULES='[{"sample_rate":1.0,"service":"xyz","name":"abc"}, \
            {"sample_rate":0.5,"service":"my-service","name":"my-name"}]'
        )
    ):
        sampling_rule = DatadogSampler().rules
        assert sampling_rule[0].sample_rate == 1.0
        assert sampling_rule[0].service == "xyz"
        assert sampling_rule[0].name == "abc"

        assert sampling_rule[1].sample_rate == 0.5
        assert sampling_rule[1].service == "my-service"
        assert sampling_rule[1].name == "my-name"
        assert len(sampling_rule) == 2

    # Testing for only Sample rate being set
    with override_env(dict(DD_TRACE_SAMPLING_RULES='[{"sample_rate":1.0}]')):
        sampling_rule = DatadogSampler().rules
        assert sampling_rule[0].sample_rate == 1.0
        assert sampling_rule[0].service == SamplingRule.NO_RULE
        assert sampling_rule[0].name == SamplingRule.NO_RULE
        assert len(sampling_rule) == 1

    # Testing for no name being set
    with override_env(dict(DD_TRACE_SAMPLING_RULES='[{"sample_rate":1.0,"service":"xyz"}]')):
        sampling_rule = DatadogSampler().rules
        assert sampling_rule[0].sample_rate == 1.0
        assert sampling_rule[0].service == "xyz"
        assert sampling_rule[0].name == SamplingRule.NO_RULE
        assert len(sampling_rule) == 1

    # Testing for no service being set
    with override_env(dict(DD_TRACE_SAMPLING_RULES='[{"sample_rate":1.0,"name":"abc"}]')):
        sampling_rule = DatadogSampler().rules
        assert sampling_rule[0].sample_rate == 1.0
        assert sampling_rule[0].service == SamplingRule.NO_RULE
        assert sampling_rule[0].name == "abc"
        assert len(sampling_rule) == 1

    # The Following error handling test use assertions on the json items instead of asserting on
    # the returned stringdue to older version of python not keeping load order in dictionaires

    # Testing for Sample rate greater than 1.0
    with pytest.raises(ValueError) as excinfo:
        with override_env(dict(DD_TRACE_SAMPLING_RULES='[{"sample_rate":2.0,"service":"xyz","name":"abc"}]')):
            sampling_rule = DatadogSampler().rules
    assert str(excinfo.value).endswith(
        "SamplingRule(sample_rate=2.0) must be greater than or equal to 0.0 and less than or equal to 1.0"
    )
    assert '"sample_rate": 2.0' in str(excinfo.value)
    assert '"service": "xyz"' in str(excinfo.value)
    assert '"name": "abc"' in str(excinfo.value)

    # Testing for no Sample rate
    with pytest.raises(KeyError) as excinfo:
        with override_env(dict(DD_TRACE_SAMPLING_RULES='[{"service":"xyz","name":"abc"}]')):
            sampling_rule = DatadogSampler().rules
    assert str(excinfo.value).startswith("'No sample_rate provided for sampling rule: ")
    assert '"service": "xyz"' in str(excinfo.value)
    assert '"name": "abc"' in str(excinfo.value)

    # Testing for Invalid JSON
    with pytest.raises(ValueError) as excinfo:
        with override_env(dict(DD_TRACE_SAMPLING_RULES='["sample_rate":1.0,"service":"xyz","name":"abc"]')):
            sampling_rule = DatadogSampler().rules
    assert 'Unable to parse DD_TRACE_SAMPLING_RULES=["sample_rate":1.0,"service":"xyz","name":"abc"]' == str(
        excinfo.value
    )

    # # Testing invalid rule with multiple rules defined
    with pytest.raises(KeyError) as excinfo:
        with override_env(
            dict(
                DD_TRACE_SAMPLING_RULES='[{"sample_rate":1.0,"service":"xyz","name":"abc"},'
                + '{"service":"my-service","name":"my-name"}]'
            )
        ):
            sampling_rule = DatadogSampler().rules
    assert str(excinfo.value).startswith("'No sample_rate provided for sampling rule: ")
    assert '"service": "my-service"' in str(excinfo.value)
    assert '"name": "my-name"' in str(excinfo.value)


@pytest.mark.parametrize(
    "span,rule,expected",
    [
        # DEV: Use sample_rate=1 to ensure SamplingRule._sample always returns True
        (create_span(name=name), SamplingRule(sample_rate=1, name=pattern), expected)
        for name, pattern, expected in [
            ("test.span", SamplingRule.NO_RULE, True),
            # DEV: `span.name` cannot be `None`
            ("test.span", None, False),
            ("test.span", "test.span", True),
            ("test.span", "test_span", False),
            ("test.span", re.compile(r"^test\.span$"), True),
            ("test_span", re.compile(r"^test.span$"), True),
            ("test.span", re.compile(r"^test_span$"), False),
            ("test.span", re.compile(r"test"), True),
            ("test.span", re.compile(r"test\.span|another\.span"), True),
            ("another.span", re.compile(r"test\.span|another\.span"), True),
            ("test.span", lambda name: "span" in name, True),
            ("test.span", lambda name: "span" not in name, False),
            ("test.span", lambda name: 1 / 0, False),
        ]
    ],
)
def test_sampling_rule_matches_name(span, rule, expected):
    assert rule.matches(span) is expected, "{} -> {} -> {}".format(rule, span, expected)


@pytest.mark.parametrize(
    "span,rule,expected",
    [
        # DEV: Use sample_rate=1 to ensure SamplingRule._sample always returns True
        (create_span(service=service), SamplingRule(sample_rate=1, service=pattern), expected)
        for service, pattern, expected in [
            ("my-service", SamplingRule.NO_RULE, True),
            ("my-service", None, False),
            (None, None, True),
            (None, "my-service", False),
            (None, re.compile(r"my-service"), False),
            (None, lambda service: "service" in service, False),
            ("my-service", "my-service", True),
            ("my-service", "my_service", False),
            ("my-service", re.compile(r"^my-"), True),
            ("my_service", re.compile(r"^my[_-]"), True),
            ("my-service", re.compile(r"^my_"), False),
            ("my-service", re.compile(r"my-service"), True),
            ("my-service", re.compile(r"my"), True),
            ("my-service", re.compile(r"my-service|another-service"), True),
            ("another-service", re.compile(r"my-service|another-service"), True),
            ("my-service", lambda service: "service" in service, True),
            ("my-service", lambda service: "service" not in service, False),
            ("my-service", lambda service: 1 / 0, False),
        ]
    ],
)
def test_sampling_rule_matches_service(span, rule, expected):
    assert rule.matches(span) is expected, "{} -> {} -> {}".format(rule, span, expected)


@pytest.mark.parametrize(
    "span,rule,expected",
    [
        # All match
        (
            create_span(
                name="test.span",
                service="my-service",
            ),
            SamplingRule(
                sample_rate=1,
                name="test.span",
                service=re.compile(r"^my-"),
            ),
            True,
        ),
        # All match,  but sample rate of 0%
        # DEV: We are checking if it is a match, not computing sampling rate, sample_rate=0 is not considered
        (
            create_span(
                name="test.span",
                service="my-service",
            ),
            SamplingRule(
                sample_rate=0,
                name="test.span",
                service=re.compile(r"^my-"),
            ),
            True,
        ),
        # Name doesn't match
        (
            create_span(
                name="test.span",
                service="my-service",
            ),
            SamplingRule(
                sample_rate=1,
                name="test_span",
                service=re.compile(r"^my-"),
            ),
            False,
        ),
        # Service doesn't match
        (
            create_span(
                name="test.span",
                service="my-service",
            ),
            SamplingRule(
                sample_rate=1,
                name="test.span",
                service=re.compile(r"^service-"),
            ),
            False,
        ),
    ],
)
def test_sampling_rule_matches(span, rule, expected):
    assert rule.matches(span) is expected, "{} -> {} -> {}".format(rule, span, expected)


def test_sampling_rule_matches_exception():
    e = Exception("an error occurred")

    def pattern(prop):
        raise e

    rule = SamplingRule(sample_rate=1.0, name=pattern)
    span = create_span(name="test.span")

    with mock.patch("ddtrace.sampler.log") as mock_log:
        assert rule.matches(span) is False
        mock_log.warning.assert_called_once_with(
            "%r pattern %r failed with %r",
            rule,
            pattern,
            "test.span",
            exc_info=True,
        )


@pytest.mark.parametrize("sample_rate", [0.01, 0.1, 0.15, 0.25, 0.5, 0.75, 0.85, 0.9, 0.95, 0.991])
def test_sampling_rule_sample(sample_rate):
    rule = SamplingRule(sample_rate=sample_rate)

    iterations = int(1e4 / sample_rate)
    sampled = sum(rule.sample(Span(name=str(i))) for i in range(iterations))

    # Less than 5% deviation when 'enough' iterations (arbitrary, just check if it converges)
    deviation = abs(sampled - (iterations * sample_rate)) / (iterations * sample_rate)
    assert deviation < 0.05, "Deviation {!r} too high with sample_rate {!r} for {} sampled".format(
        deviation, sample_rate, sampled
    )


def test_sampling_rule_sample_rate_1():
    rule = SamplingRule(sample_rate=1)

    iterations = int(1e4)
    assert all(rule.sample(Span(name=str(i))) for i in range(iterations))


def test_sampling_rule_sample_rate_0():
    rule = SamplingRule(sample_rate=0)

    iterations = int(1e4)
    assert sum(rule.sample(Span(name=str(i))) for i in range(iterations)) == 0


def test_datadog_sampler_init():
    # No args
    sampler = DatadogSampler()
    assert sampler.rules == []
    assert isinstance(sampler.limiter, RateLimiter)
    assert sampler.limiter.rate_limit == DatadogSampler.DEFAULT_RATE_LIMIT

    # With rules
    rule = SamplingRule(sample_rate=1)
    sampler = DatadogSampler(rules=[rule])
    assert sampler.rules == [rule]
    assert sampler.limiter.rate_limit == DatadogSampler.DEFAULT_RATE_LIMIT

    # With rate limit
    sampler = DatadogSampler(rate_limit=10)
    assert sampler.limiter.rate_limit == 10

    # With default_sample_rate
    sampler = DatadogSampler(default_sample_rate=0.5)
    assert sampler.limiter.rate_limit == DatadogSampler.DEFAULT_RATE_LIMIT
    assert sampler.rules == [SamplingRule(sample_rate=0.5)]

    # From env variables
    with override_env(dict(DD_TRACE_SAMPLE_RATE="0.5", DD_TRACE_RATE_LIMIT="10")):
        sampler = DatadogSampler()
        assert sampler.limiter.rate_limit == 10
        assert sampler.rules == [SamplingRule(sample_rate=0.5)]

    # DD_TRACE_SAMPLE_RATE=0
    with override_env(dict(DD_TRACE_SAMPLE_RATE="0")):
        sampler = DatadogSampler()
        assert sampler.limiter.rate_limit == DatadogSampler.DEFAULT_RATE_LIMIT
        assert sampler.rules == [SamplingRule(sample_rate=0)]

    # Invalid env vars
    with override_env(dict(DD_TRACE_SAMPLE_RATE="asdf")):
        with pytest.raises(ValueError):
            DatadogSampler()

    # Invalid env vars
    with override_env(dict(DD_TRACE_RATE_LIMIT="asdf")):
        with pytest.raises(ValueError):
            DatadogSampler()

    # Invalid rules
    for val in (None, True, False, object(), 1, Exception()):
        with pytest.raises(TypeError):
            DatadogSampler(rules=[val])

    # Ensure rule order
    rule_1 = SamplingRule(sample_rate=1)
    rule_2 = SamplingRule(sample_rate=0.5, service="test")
    rule_3 = SamplingRule(sample_rate=0.25, name="flask.request")
    sampler = DatadogSampler(rules=[rule_1, rule_2, rule_3])
    assert sampler.rules == [rule_1, rule_2, rule_3]

    # Ensure default rule is appended
    sampler = DatadogSampler(rules=[rule_1, rule_2, rule_3], default_sample_rate=0.75)
    assert sampler.rules == [rule_1, rule_2, rule_3, SamplingRule(sample_rate=0.75)]


@mock.patch("ddtrace.sampler.RateSampler.sample")
def test_datadog_sampler_sample_no_rules(mock_sample, dummy_tracer):
    sampler = DatadogSampler()
    dummy_tracer.configure(sampler=sampler)

    # Default RateByServiceSampler() is applied
    #   No rules configured
    #   No global rate limit
    #   No rate limit configured
    # RateByServiceSampler.sample(span) returns True
    mock_sample.return_value = True
    with dummy_tracer.trace("test"):
        pass
    spans = dummy_tracer.pop()
    assert len(spans) == 1, "Span should have been written"
    assert_sampling_decision_tags(
        spans[0],
        agent=1.0,
        limit=None,
        rule=None,
        sampling_priority=AUTO_KEEP,
        # default sampling decision
        trace_tag="-0",
    )

    # Default RateByServiceSampler() is applied
    #   No rules configured
    #   No global rate limit
    #   No rate limit configured
    # RateByServiceSampler.sample(span) returns False
    mock_sample.return_value = False
    with dummy_tracer.trace("test"):
        pass
    spans = dummy_tracer.pop()
    assert len(spans) == 1, "Span should have been written"
    assert_sampling_decision_tags(
        spans[0], agent=1.0, limit=None, rule=None, sampling_priority=AUTO_REJECT, trace_tag=None
    )


class MatchSample(SamplingRule):
    def matches(self, span):
        return True

    def sample(self, span):
        return True


class NoMatch(SamplingRule):
    def matches(self, span):
        return False

    def sample(self, span):
        return True


class MatchNoSample(SamplingRule):
    def matches(self, span):
        return True

    def sample(self, span):
        return False


@pytest.mark.parametrize(
    "sampler, sampling_priority, sampling_mechanism, rule, limit",
    [
        (
            DatadogSampler(
                default_sample_rate=1.0,
                rules=[
                    NoMatch(0.5),
                    NoMatch(0.5),
                    NoMatch(0.5),
                ],
            ),
            USER_KEEP,
            SamplingMechanism.TRACE_SAMPLING_RULE,
            1.0,
            None,
        ),
        (
            DatadogSampler(
                default_sample_rate=1.0,
                rules=[
                    NoMatch(0.5),
                    NoMatch(0.5),
                    MatchSample(0.5),
                ],
            ),
            USER_KEEP,
            SamplingMechanism.TRACE_SAMPLING_RULE,
            0.5,
            None,
        ),
        (
            DatadogSampler(
                default_sample_rate=1.0,
                rules=[
                    MatchSample(0.5),
                    MatchNoSample(0.5),
                    MatchNoSample(0.5),
                ],
            ),
            USER_KEEP,
            SamplingMechanism.TRACE_SAMPLING_RULE,
            0.5,
            None,
        ),
        (
            DatadogSampler(
                default_sample_rate=1.0,
                rules=[
                    NoMatch(0.5),
                    MatchNoSample(0.5),
                    NoMatch(0.5),
                ],
            ),
            USER_REJECT,
            None,
            0.5,
            None,
        ),
        (
            DatadogSampler(
                default_sample_rate=1.0,
                rules=[
                    NoMatch(0.5),
                    MatchNoSample(0.5),
                    NoMatch(0.5),
                ],
            ),
            USER_REJECT,
            None,
            0.5,
            None,
        ),
        (
            DatadogSampler(
                default_sample_rate=0,
            ),
            USER_REJECT,
            None,
            0,
            None,
        ),
        (
            DatadogSampler(
                default_sample_rate=1.0,
                rate_limit=0,
            ),
            USER_REJECT,
            None,
            1.0,
            0.0,
        ),
    ],
)
def test_datadog_sampler_sample_rules(sampler, sampling_priority, sampling_mechanism, rule, limit, dummy_tracer):
    dummy_tracer.configure(sampler=sampler)
    with dummy_tracer.trace("span"):
        pass
    spans = dummy_tracer.pop()

    # A span is always expected as DatadogSamplers are used.
    assert len(spans) > 0
    span = spans[0]
    # Hence, it should always be "sampled" (sent to the agent)
    # This is an implementation detail so we probably don't have to
    # test it.
    assert span.sampled
    trace_tag = "-%d" % sampling_mechanism if sampling_mechanism is not None else None
    assert_sampling_decision_tags(
        span, rule=rule, limit=limit, sampling_priority=sampling_priority, trace_tag=trace_tag
    )


def test_datadog_sampler_tracer(dummy_tracer):
    rule = SamplingRule(sample_rate=1.0, name="test.span")
    sampler = DatadogSampler(rules=[rule])
    dummy_tracer.configure(sampler=sampler)

    with dummy_tracer.trace("test.span"):
        pass

    spans = dummy_tracer.pop()
    assert len(spans) == 1, "Span should have been sampled and written"
    assert_sampling_decision_tags(
        spans[0],
        rule=1.0,
        limit=None,
        sampling_priority=USER_KEEP,
        # sampling decision from user sampling rule
        trace_tag="-3",
    )


def test_datadog_sampler_tracer_rate_limited(dummy_tracer):
    rule = SamplingRule(sample_rate=1.0, name="test.span")
    sampler = DatadogSampler(rules=[rule], rate_limit=0)
    dummy_tracer.configure(sampler=sampler)

    with dummy_tracer.trace("test.span"):
        pass

    spans = dummy_tracer.pop()
    assert len(spans) == 1, "Span should have been sampled and written"
    assert_sampling_decision_tags(spans[0], rule=1.0, limit=0.0, sampling_priority=USER_REJECT, trace_tag=None)


def test_datadog_sampler_tracer_rate_0(dummy_tracer):
    # Sample rate of 0 means never sample
    rule = SamplingRule(sample_rate=0, name="test.span")
    sampler = DatadogSampler(rules=[rule])
    dummy_tracer.configure(sampler=sampler)

    with dummy_tracer.trace("test.span"):
        pass

    spans = dummy_tracer.pop()
    assert len(spans) == 1, "Span should have been sampled and written"
    assert_sampling_decision_tags(spans[0], rule=0.0, sampling_priority=USER_REJECT)


def test_datadog_sampler_tracer_child(dummy_tracer):
    # No rules means it gets applied to every span
    rule = SamplingRule(sample_rate=1.0)
    sampler = DatadogSampler(rules=[rule])
    dummy_tracer.configure(sampler=sampler)

    with dummy_tracer.trace("parent.span"):
        with dummy_tracer.trace("child.span"):
            pass

    spans = dummy_tracer.pop()
    assert len(spans) == 2, "Trace should have been sampled and written"
    assert_sampling_decision_tags(
        spans[0],
        rule=1.0,
        limit=None,
        sampling_priority=USER_KEEP,
        # sampling decision from user sampling rule
        trace_tag="-3",
    )
    assert_sampling_decision_tags(
        spans[1],
        agent=None,
        rule=None,
        limit=None,
        # DEV: the trace tag check is on the context which is shared between parent and child
        trace_tag="-3",
    )


def test_datadog_sampler_tracer_start_span(dummy_tracer):
    # No rules means it gets applied to every span
    rule = SamplingRule(sample_rate=1.0)
    sampler = DatadogSampler(rules=[rule])
    dummy_tracer.configure(sampler=sampler)

    span = dummy_tracer.start_span("test.span")
    span.finish()

    spans = dummy_tracer.pop()
    assert len(spans) == 1, "Span should have been sampled and written"
    assert_sampling_decision_tags(
        spans[0],
        rule=1.0,
        limit=None,
        sampling_priority=USER_KEEP,
        # sampling decision from user sampling rule
        trace_tag="-3",
    )


def test_datadog_sampler_update_rate_by_service_sample_rates(dummy_tracer):
    cases = [
        {
            "service:,env:": 1,
        },
        {
            "service:,env:": 1,
            "service:mcnulty,env:dev": 0.33,
            "service:postgres,env:dev": 0.7,
        },
        {
            "service:,env:": 1,
            "service:mcnulty,env:dev": 0.25,
            "service:postgres,env:dev": 0.5,
            "service:redis,env:prod": 0.75,
        },
    ]

    # By default sampler sets it's default sampler to RateByServiceSampler
    sampler = DatadogSampler()
    for case in cases:
        sampler.update_rate_by_service_sample_rates(case)
        rates = {}
        for k, v in iteritems(sampler._by_service_samplers):
            rates[k] = v.sample_rate
        assert case == rates, "%s != %s" % (case, rates)

    # It's important to also test in reverse mode for we want to make sure key deletion
    # works as well as key insertion (and doing this both ways ensures we trigger both cases)
    cases.reverse()
    for case in cases:
        sampler.update_rate_by_service_sample_rates(case)
        rates = {}
        for k, v in iteritems(sampler._by_service_samplers):
            rates[k] = v.sample_rate
        assert case == rates, "%s != %s" % (case, rates)



from typing import Dict
from typing import FrozenSet
from typing import Optional
from typing import Text
from typing import cast

from ddtrace import config

from ..constants import AUTO_KEEP
from ..constants import AUTO_REJECT
from ..constants import USER_KEEP
from ..context import Context
from ..internal._tagset import TagsetDecodeError
from ..internal._tagset import TagsetEncodeError
from ..internal._tagset import TagsetMaxSizeDecodeError
from ..internal._tagset import TagsetMaxSizeEncodeError
from ..internal._tagset import decode_tagset_string
from ..internal._tagset import encode_tagset_values
from ..internal.compat import ensure_str
from ..internal.compat import ensure_text
from ..internal.constants import PROPAGATION_STYLE_B3
from ..internal.constants import PROPAGATION_STYLE_B3_SINGLE_HEADER
from ..internal.constants import PROPAGATION_STYLE_DATADOG
from ..internal.logger import get_logger
from ..internal.sampling import validate_sampling_decision
from ..span import _MetaDictType
from ._utils import get_wsgi_header


log = get_logger(__name__)

# HTTP headers one should set for distributed tracing.
# These are cross-language (eg: Python, Go and other implementations should honor these)
HTTP_HEADER_TRACE_ID = "x-datadog-trace-id"
HTTP_HEADER_PARENT_ID = "x-datadog-parent-id"
HTTP_HEADER_SAMPLING_PRIORITY = "x-datadog-sampling-priority"
HTTP_HEADER_ORIGIN = "x-datadog-origin"
_HTTP_HEADER_B3_SINGLE = "b3"
_HTTP_HEADER_B3_TRACE_ID = "x-b3-traceid"
_HTTP_HEADER_B3_SPAN_ID = "x-b3-spanid"
_HTTP_HEADER_B3_SAMPLED = "x-b3-sampled"
_HTTP_HEADER_B3_FLAGS = "x-b3-flags"
_HTTP_HEADER_TAGS = "x-datadog-tags"


def _possible_header(header):
    # type: (str) -> FrozenSet[str]
    return frozenset([header, get_wsgi_header(header).lower()])


# Note that due to WSGI spec we have to also check for uppercased and prefixed
# versions of these headers
POSSIBLE_HTTP_HEADER_TRACE_IDS = _possible_header(HTTP_HEADER_TRACE_ID)
POSSIBLE_HTTP_HEADER_PARENT_IDS = _possible_header(HTTP_HEADER_PARENT_ID)
POSSIBLE_HTTP_HEADER_SAMPLING_PRIORITIES = _possible_header(HTTP_HEADER_SAMPLING_PRIORITY)
POSSIBLE_HTTP_HEADER_ORIGIN = _possible_header(HTTP_HEADER_ORIGIN)
_POSSIBLE_HTTP_HEADER_TAGS = frozenset([_HTTP_HEADER_TAGS, get_wsgi_header(_HTTP_HEADER_TAGS).lower()])
_POSSIBLE_HTTP_HEADER_B3_SINGLE_HEADER = _possible_header(_HTTP_HEADER_B3_SINGLE)
_POSSIBLE_HTTP_HEADER_B3_TRACE_IDS = _possible_header(_HTTP_HEADER_B3_TRACE_ID)
_POSSIBLE_HTTP_HEADER_B3_SPAN_IDS = _possible_header(_HTTP_HEADER_B3_SPAN_ID)
_POSSIBLE_HTTP_HEADER_B3_SAMPLEDS = _possible_header(_HTTP_HEADER_B3_SAMPLED)
_POSSIBLE_HTTP_HEADER_B3_FLAGS = _possible_header(_HTTP_HEADER_B3_FLAGS)


def _extract_header_value(possible_header_names, headers, default=None):
    # type: (FrozenSet[str], Dict[str, str], Optional[str]) -> Optional[str]
    for header in possible_header_names:
        if header in headers:
            return ensure_str(headers[header], errors="backslashreplace")

    return default


def _b3_id_to_dd_id(b3_id):
    # type: (str) -> int
    """Helper to convert B3 trace/span hex ids into Datadog compatible ints

    If the id is > 64 bit then truncate the trailing 64 bit.

    "463ac35c9f6413ad48485a3953bb6124" -> "48485a3953bb6124" -> 5208512171318403364
    """
    return int(b3_id[-16:], 16)


def _dd_id_to_b3_id(dd_id):
    # type: (int) -> str
    """Helper to convert Datadog trace/span int ids into B3 compatible hex ids"""
    # DEV: `hex(dd_id)` will give us `0xDEADBEEF`
    # DEV: this gives us lowercase hex, which is what we want
    return "{:x}".format(dd_id)


class _DatadogMultiHeader:
    """Helper class for injecting/extract Datadog multi header format

    Headers:

      - ``x-datadog-trace-id`` the context trace id as a uint64 integer
      - ``x-datadog-parent-id`` the context current span id as a uint64 integer
      - ``x-datadog-sampling-priority`` integer representing the sampling decision.
        ``<= 0`` (Reject) or ``> 1`` (Keep)
      - ``x-datadog-origin`` optional name of origin Datadog product which initiated the request
      - ``x-datadog-tags`` optional tracer tags

    Restrictions:

      - Trace tag key-value pairs in ``x-datadog-tags`` are extracted from incoming requests.
      - Only trace tags with keys prefixed with ``_dd.p.`` are propagated.
      - The trace tag keys must be printable ASCII characters excluding space, comma, and equals.
      - The trace tag values must be printable ASCII characters excluding comma. Leading and
        trailing spaces are trimmed.
    """

    _X_DATADOG_TAGS_EXTRACT_REJECT = frozenset(["_dd.p.upstream_services"])

    @staticmethod
    def _is_valid_datadog_trace_tag_key(key):
        return key.startswith("_dd.p.")

    @staticmethod
    def _inject(span_context, headers):
        # type: (Context, Dict[str, str]) -> None
        if span_context.trace_id is None or span_context.span_id is None:
            log.debug("tried to inject invalid context %r", span_context)
            return

        headers[HTTP_HEADER_TRACE_ID] = str(span_context.trace_id)
        headers[HTTP_HEADER_PARENT_ID] = str(span_context.span_id)
        sampling_priority = span_context.sampling_priority
        # Propagate priority only if defined
        if sampling_priority is not None:
            headers[HTTP_HEADER_SAMPLING_PRIORITY] = str(span_context.sampling_priority)
        # Propagate origin only if defined
        if span_context.dd_origin is not None:
            headers[HTTP_HEADER_ORIGIN] = ensure_text(span_context.dd_origin)

        if not config._x_datadog_tags_enabled:
            span_context._meta["_dd.propagation_error"] = "disabled"
            return

        # Do not try to encode tags if we have already tried and received an error
        if "_dd.propagation_error" in span_context._meta:
            return

        # Only propagate trace tags which means ignoring the _dd.origin
        tags_to_encode = {
            # DEV: Context._meta is a _MetaDictType but we need Dict[str, str]
            ensure_str(k): ensure_str(v)
            for k, v in span_context._meta.items()
            if _DatadogMultiHeader._is_valid_datadog_trace_tag_key(k)
        }  # type: Dict[Text, Text]

        if tags_to_encode:
            try:
                headers[_HTTP_HEADER_TAGS] = encode_tagset_values(
                    tags_to_encode, max_size=config._x_datadog_tags_max_length
                )
            except TagsetMaxSizeEncodeError:
                # We hit the max size allowed, add a tag to the context to indicate this happened
                span_context._meta["_dd.propagation_error"] = "inject_max_size"
                log.warning("failed to encode x-datadog-tags", exc_info=True)
            except TagsetEncodeError:
                # We hit an encoding error, add a tag to the context to indicate this happened
                span_context._meta["_dd.propagation_error"] = "encoding_error"
                log.warning("failed to encode x-datadog-tags", exc_info=True)

    @staticmethod
    def _extract(headers):
        # type: (Dict[str, str]) -> Optional[Context]
        trace_id = _extract_header_value(
            POSSIBLE_HTTP_HEADER_TRACE_IDS,
            headers,
        )
        if trace_id is None:
            return None

        parent_span_id = _extract_header_value(
            POSSIBLE_HTTP_HEADER_PARENT_IDS,
            headers,
            default="0",
        )
        sampling_priority = _extract_header_value(
            POSSIBLE_HTTP_HEADER_SAMPLING_PRIORITIES,
            headers,
        )
        origin = _extract_header_value(
            POSSIBLE_HTTP_HEADER_ORIGIN,
            headers,
        )

        meta = None
        tags_value = _extract_header_value(
            _POSSIBLE_HTTP_HEADER_TAGS,
            headers,
            default="",
        )
        if tags_value:
            # Do not fail if the tags are malformed
            try:
                meta = {
                    k: v
                    for (k, v) in decode_tagset_string(tags_value).items()
                    if (
                        k not in _DatadogMultiHeader._X_DATADOG_TAGS_EXTRACT_REJECT
                        and _DatadogMultiHeader._is_valid_datadog_trace_tag_key(k)
                    )
                }
            except TagsetMaxSizeDecodeError:
                meta = {
                    "_dd.propagation_error": "extract_max_size",
                }
                log.warning("failed to decode x-datadog-tags", exc_info=True)
            except TagsetDecodeError:
                meta = {
                    "_dd.propagation_error": "decoding_error",
                }
                log.debug("failed to decode x-datadog-tags: %r", tags_value, exc_info=True)

        # Try to parse values into their expected types
        try:
            if sampling_priority is not None:
                sampling_priority = int(sampling_priority)  # type: ignore[assignment]
            else:
                sampling_priority = sampling_priority

            if meta:
                meta = validate_sampling_decision(meta)

            return Context(
                # DEV: Do not allow `0` for trace id or span id, use None instead
                trace_id=int(trace_id) or None,
                span_id=int(parent_span_id) or None,  # type: ignore[arg-type]
                sampling_priority=sampling_priority,  # type: ignore[arg-type]
                dd_origin=origin,
                # DEV: This cast is needed because of the type requirements of
                # span tags and trace tags which are currently implemented using
                # the same type internally (_MetaDictType).
                meta=cast(_MetaDictType, meta),
            )
        except (TypeError, ValueError):
            log.debug(
                (
                    "received invalid x-datadog-* headers, "
                    "trace-id: %r, parent-id: %r, priority: %r, origin: %r, tags:%r"
                ),
                trace_id,
                parent_span_id,
                sampling_priority,
                origin,
                tags_value,
            )
        return None


class _B3MultiHeader:
    """Helper class to inject/extract B3 Multi-Headers

    https://github.com/openzipkin/b3-propagation/blob/3e54cda11620a773d53c7f64d2ebb10d3a01794c/README.md#multiple-headers

    Example::

        X-B3-TraceId: 80f198ee56343ba864fe8b2a57d3eff7
        X-B3-ParentSpanId: 05e3ac9a4f6e3b90
        X-B3-SpanId: e457b5a2e4d86bd1
        X-B3-Sampled: 1


    Headers:

      - ``X-B3-TraceId`` header is encoded as 32 or 16 lower-hex characters.
      - ``X-B3-SpanId`` header is encoded as 16 lower-hex characters.
      - ``X-B3-Sampled`` header value of ``0`` means Deny, ``1`` means Accept, and absent means to defer.
      - ``X-B3-Flags`` header is used to set ``1`` meaning Debug or an Accept.

    Restrictions:

      - ``X-B3-Sampled`` and ``X-B3-Flags`` should never both be set

    Implementation details:

      - Sampling priority gets encoded as:
        - ``sampling_priority <= 0`` -> ``X-B3-Sampled: 0``
        - ``sampling_priority == 1`` -> ``X-B3-Sampled: 1``
        - ``sampling_priority > 1`` -> ``X-B3-Flags: 1``
      - Sampling priority gets decoded as:
        - ``X-B3-Sampled: 0`` -> ``sampling_priority = 0``
        - ``X-B3-Sampled: 1`` -> ``sampling_priority = 1``
        - ``X-B3-Flags: 1`` -> ``sampling_priority = 2``
      - ``X-B3-TraceId`` is not required, will use ``None`` when not present
      - ``X-B3-SpanId`` is not required, will use ``None`` when not present
    """

    @staticmethod
    def _inject(span_context, headers):
        # type: (Context, Dict[str, str]) -> None
        if span_context.trace_id is None or span_context.span_id is None:
            log.debug("tried to inject invalid context %r", span_context)
            return

        headers[_HTTP_HEADER_B3_TRACE_ID] = _dd_id_to_b3_id(span_context.trace_id)
        headers[_HTTP_HEADER_B3_SPAN_ID] = _dd_id_to_b3_id(span_context.span_id)
        sampling_priority = span_context.sampling_priority
        # Propagate priority only if defined
        if sampling_priority is not None:
            if sampling_priority <= 0:
                headers[_HTTP_HEADER_B3_SAMPLED] = "0"
            elif sampling_priority == 1:
                headers[_HTTP_HEADER_B3_SAMPLED] = "1"
            elif sampling_priority > 1:
                headers[_HTTP_HEADER_B3_FLAGS] = "1"

    @staticmethod
    def _extract(headers):
        # type: (Dict[str, str]) -> Optional[Context]
        trace_id_val = _extract_header_value(
            _POSSIBLE_HTTP_HEADER_B3_TRACE_IDS,
            headers,
        )
        if trace_id_val is None:
            return None

        span_id_val = _extract_header_value(
            _POSSIBLE_HTTP_HEADER_B3_SPAN_IDS,
            headers,
        )
        sampled = _extract_header_value(
            _POSSIBLE_HTTP_HEADER_B3_SAMPLEDS,
            headers,
        )
        flags = _extract_header_value(
            _POSSIBLE_HTTP_HEADER_B3_FLAGS,
            headers,
        )

        # Try to parse values into their expected types
        try:
            # DEV: We are allowed to have only x-b3-sampled/flags
            # DEV: Do not allow `0` for trace id or span id, use None instead
            trace_id = None
            span_id = None
            if trace_id_val is not None:
                trace_id = _b3_id_to_dd_id(trace_id_val) or None
            if span_id_val is not None:
                span_id = _b3_id_to_dd_id(span_id_val) or None

            sampling_priority = None
            if sampled is not None:
                if sampled == "0":
                    sampling_priority = AUTO_REJECT
                elif sampled == "1":
                    sampling_priority = AUTO_KEEP
            if flags == "1":
                sampling_priority = USER_KEEP

            return Context(
                trace_id=trace_id,
                span_id=span_id,
                sampling_priority=sampling_priority,
            )
        except (TypeError, ValueError):
            log.debug(
                "received invalid x-b3-* headers, " "trace-id: %r, span-id: %r, sampled: %r, flags: %r",
                trace_id_val,
                span_id_val,
                sampled,
                flags,
            )
        return None


class _B3SingleHeader:
    """Helper class to inject/extract B3 Single Header

    https://github.com/openzipkin/b3-propagation/blob/3e54cda11620a773d53c7f64d2ebb10d3a01794c/README.md#single-header

    Format::

        b3={TraceId}-{SpanId}-{SamplingState}-{ParentSpanId}

    Example::

        b3: 80f198ee56343ba864fe8b2a57d3eff7-e457b5a2e4d86bd1-1-05e3ac9a4f6e3b90


    Values:

      - ``TraceId`` header is encoded as 32 or 16 lower-hex characters.
      - ``SpanId`` header is encoded as 16 lower-hex characters.
      - ``SamplingState`` header value of ``0`` means Deny, ``1`` means Accept, and ``d`` means Debug
      - ``ParentSpanId`` header is not used/ignored if sent

    Restrictions:

      - ``ParentSpanId`` value is ignored/not used

    Implementation details:

      - Sampling priority gets encoded as:
        - ``sampling_priority <= 0`` -> ``SamplingState: 0``
        - ``sampling_priority == 1`` -> ``SamplingState: 1``
        - ``sampling_priority > 1`` -> ``SamplingState: d``
      - Sampling priority gets decoded as:
        - ``SamplingState: 0`` -> ``sampling_priority = 0``
        - ``SamplingState: 1`` -> ``sampling_priority = 1``
        - ``SamplingState: d`` -> ``sampling_priority = 2``
      - ``TraceId`` is not required, will use ``None`` when not present
      - ``SpanId`` is not required, will use ``None`` when not present
    """

    @staticmethod
    def _inject(span_context, headers):
        # type: (Context, Dict[str, str]) -> None
        if span_context.trace_id is None or span_context.span_id is None:
            log.debug("tried to inject invalid context %r", span_context)
            return

        single_header = "{}-{}".format(_dd_id_to_b3_id(span_context.trace_id), _dd_id_to_b3_id(span_context.span_id))
        sampling_priority = span_context.sampling_priority
        if sampling_priority is not None:
            if sampling_priority <= 0:
                single_header += "-0"
            elif sampling_priority == 1:
                single_header += "-1"
            elif sampling_priority > 1:
                single_header += "-d"
        headers[_HTTP_HEADER_B3_SINGLE] = single_header

    @staticmethod
    def _extract(headers):
        # type: (Dict[str, str]) -> Optional[Context]
        single_header = _extract_header_value(_POSSIBLE_HTTP_HEADER_B3_SINGLE_HEADER, headers)
        if not single_header:
            return None

        trace_id = None
        span_id = None
        sampled = None

        parts = single_header.split("-")
        trace_id_val = None
        span_id_val = None

        # Only SamplingState is provided
        if len(parts) == 1:
            (sampled,) = parts

        # Only TraceId and SpanId are provided
        elif len(parts) == 2:
            trace_id_val, span_id_val = parts

        # Full header, ignore any ParentSpanId present
        elif len(parts) >= 3:
            trace_id_val, span_id_val, sampled = parts[:3]

        # Try to parse values into their expected types
        try:
            # DEV: We are allowed to have only x-b3-sampled/flags
            # DEV: Do not allow `0` for trace id or span id, use None instead
            if trace_id_val is not None:
                trace_id = _b3_id_to_dd_id(trace_id_val) or None
            if span_id_val is not None:
                span_id = _b3_id_to_dd_id(span_id_val) or None

            sampling_priority = None
            if sampled is not None:
                if sampled == "0":
                    sampling_priority = AUTO_REJECT
                elif sampled == "1":
                    sampling_priority = AUTO_KEEP
                elif sampled == "d":
                    sampling_priority = USER_KEEP

            return Context(
                trace_id=trace_id,
                span_id=span_id,
                sampling_priority=sampling_priority,
            )
        except (TypeError, ValueError):
            log.debug(
                "received invalid b3 header, b3: %r",
                single_header,
            )
        return None


class HTTPPropagator(object):
    """A HTTP Propagator using HTTP headers as carrier."""

    @staticmethod
    def inject(span_context, headers):
        # type: (Context, Dict[str, str]) -> None
        """Inject Context attributes that have to be propagated as HTTP headers.

        Here is an example using `requests`::

            import requests
            from ddtrace.propagation.http import HTTPPropagator

            def parent_call():
                with tracer.trace('parent_span') as span:
                    headers = {}
                    HTTPPropagator.inject(span.context, headers)
                    url = '<some RPC endpoint>'
                    r = requests.get(url, headers=headers)

        :param Context span_context: Span context to propagate.
        :param dict headers: HTTP headers to extend with tracing attributes.
        """
        # Not a valid context to propagate
        if span_context.trace_id is None or span_context.span_id is None:
            log.debug("tried to inject invalid context %r", span_context)
            return

        if PROPAGATION_STYLE_DATADOG in config._propagation_style_inject:
            _DatadogMultiHeader._inject(span_context, headers)
        if PROPAGATION_STYLE_B3 in config._propagation_style_inject:
            _B3MultiHeader._inject(span_context, headers)
        if PROPAGATION_STYLE_B3_SINGLE_HEADER in config._propagation_style_inject:
            _B3SingleHeader._inject(span_context, headers)

    @staticmethod
    def extract(headers):
        # type: (Dict[str,str]) -> Context
        """Extract a Context from HTTP headers into a new Context.

        Here is an example from a web endpoint::

            from ddtrace.propagation.http import HTTPPropagator

            def my_controller(url, headers):
                context = HTTPPropagator.extract(headers)
                if context:
                    tracer.context_provider.activate(context)

                with tracer.trace('my_controller') as span:
                    span.set_tag('http.url', url)

        :param dict headers: HTTP headers to extract tracing attributes.
        :return: New `Context` with propagated attributes.
        """
        if not headers:
            return Context()

        try:
            normalized_headers = {name.lower(): v for name, v in headers.items()}
            # Check all styles until we find the first valid match
            # DEV: We want to check them in this specific priority order
            if PROPAGATION_STYLE_DATADOG in config._propagation_style_extract:
                context = _DatadogMultiHeader._extract(normalized_headers)
                if context is not None:
                    return context
            if PROPAGATION_STYLE_B3 in config._propagation_style_extract:
                context = _B3MultiHeader._extract(normalized_headers)
                if context is not None:
                    return context
            if PROPAGATION_STYLE_B3_SINGLE_HEADER in config._propagation_style_extract:
                context = _B3SingleHeader._extract(normalized_headers)
                if context is not None:
                    return context
        except Exception:
            log.debug("error while extracting context propagation headers", exc_info=True)
        return Context()



import functools
import importlib
import sys
import unittest

from ddtrace.vendor import wrapt
from tests.subprocesstest import SubprocessTestCase
from tests.subprocesstest import run_in_subprocess


class PatchMixin(unittest.TestCase):
    """
    TestCase for testing the patch logic of an integration.
    """

    def module_imported(self, modname):
        """
        Returns whether a module is imported or not.
        """
        return modname in sys.modules

    def assert_module_imported(self, modname):
        """
        Asserts that the module, given its name is imported.
        """
        assert self.module_imported(modname), "{} module not imported".format(modname)

    def assert_not_module_imported(self, modname):
        """
        Asserts that the module, given its name is not imported.
        """
        assert not self.module_imported(modname), "{} module is imported".format(modname)

    def is_wrapped(self, obj):
        return isinstance(obj, wrapt.ObjectProxy)

    def assert_wrapped(self, obj):
        """
        Helper to assert that a given object is properly wrapped by wrapt.
        """
        self.assertTrue(self.is_wrapped(obj), "{} is not wrapped".format(obj))

    def assert_not_wrapped(self, obj):
        """
        Helper to assert that a given object is not wrapped by wrapt.
        """
        self.assertFalse(self.is_wrapped(obj), "{} is wrapped".format(obj))

    def assert_not_double_wrapped(self, obj):
        """
        Helper to assert that a given already wrapped object is not wrapped twice.

        This is useful for asserting idempotence.
        """
        self.assert_wrapped(obj)
        self.assert_not_wrapped(obj.__wrapped__)


def raise_if_no_attrs(f):
    """
    A helper for PatchTestCase test methods that will check if there are any
    modules to use else raise a NotImplementedError.

    :param f: method to wrap with a check
    """
    required_attrs = [
        "__module_name__",
        "__integration_name__",
        "__patch_func__",
    ]

    @functools.wraps(f)
    def checked_method(self, *args, **kwargs):
        for attr in required_attrs:
            if not getattr(self, attr):
                raise NotImplementedError(f.__doc__)
        return f(self, *args, **kwargs)

    return checked_method


def noop_if_no_unpatch(f):
    """
    A helper for PatchTestCase test methods that will no-op the test if the
    __unpatch_func__ attribute is None
    """

    @functools.wraps(f)
    def wrapper(self, *args, **kwargs):
        if getattr(self, "__unpatch_func__") is None:
            self.skipTest(reason="No unpatch function given")
        return f(self, *args, **kwargs)

    return wrapper


class PatchTestCase(object):
    """
    unittest or other test runners will pick up the base test case as a testcase
    since it inherits from unittest.TestCase unless we wrap it with this empty
    parent class.
    """

    @run_in_subprocess
    class Base(SubprocessTestCase, PatchMixin):
        """Provides default test methods to be used for testing common integration patching logic.
        Each test method provides a default implementation which will use the
        provided attributes (described below). If the attributes are not
        provided a NotImplementedError will be raised for each method that is
        not overridden.

        Attributes:
            __integration_name__ the name of the integration.
            __module_name__ module which the integration patches.
            __patch_func__ patch function from the integration.
            __unpatch_func__ unpatch function from the integration.

        Example:
        A simple implementation inheriting this TestCase looks like::

            from ddtrace.contrib.redis import patch, unpatch

            class RedisPatchTestCase(PatchTestCase.Base):
                __integration_name__ = 'redis'
                __module_name__ 'redis'
                __patch_func__ = patch
                __unpatch_func__ = unpatch

                def assert_module_patched(self, redis):
                    # assert patching logic
                    # self.assert_wrapped(...)

                def assert_not_module_patched(self, redis):
                    # assert patching logic
                    # self.assert_not_wrapped(...)

                def assert_not_module_double_patched(self, redis):
                    # assert patching logic
                    # self.assert_not_double_wrapped(...)

                # override this particular test case
                def test_patch_import(self):
                    # custom patch before import check

                # optionally override other test methods...
        """

        __integration_name__ = None
        __module_name__ = None
        __patch_func__ = None
        __unpatch_func__ = None

        def __init__(self, *args, **kwargs):
            # DEV: Python will wrap a function when assigning it to a class as an
            # attribute. So we cannot call self.__unpatch_func__() because the `self`
            # reference will be passed as an argument.
            # So we need to unwrap the function and then wrap it in a function
            # that will absorb the unpatch function.
            if self.__unpatch_func__:
                unpatch_func = self.__unpatch_func__.__func__

                def unpatch():
                    unpatch_func()

                self.__unpatch_func__ = unpatch

            # Same for __patch_func__()
            if self.__patch_func__:
                patch_func = self.__patch_func__.__func__

                def patch():
                    patch_func()

                self.__patch_func__ = patch
            super(PatchTestCase.Base, self).__init__(*args, **kwargs)

        def _gen_test_attrs(self, ops):
            """
            A helper to return test names for tests given a list of different
            operations.
            :return:
            """
            from itertools import permutations

            return ["test_{}".format("_".join(c)) for c in permutations(ops, len(ops))]

        def test_verify_test_coverage(self):
            """
            This TestCase should cover a variety of combinations of importing,
            patching and unpatching.
            """
            tests = []
            tests += self._gen_test_attrs(["import", "patch"])
            tests += self._gen_test_attrs(["import", "patch", "patch"])
            tests += self._gen_test_attrs(["import", "patch", "unpatch"])
            tests += self._gen_test_attrs(["import", "patch", "unpatch", "unpatch"])

            # TODO: it may be possible to generate test cases dynamically. For
            # now focus on the important ones.
            test_ignore = set(
                [
                    "test_unpatch_import_patch",
                    "test_import_unpatch_patch_unpatch",
                    "test_import_unpatch_unpatch_patch",
                    "test_patch_import_unpatch_unpatch",
                    "test_unpatch_import_patch_unpatch",
                    "test_unpatch_import_unpatch_patch",
                    "test_unpatch_patch_import_unpatch",
                    "test_unpatch_patch_unpatch_import",
                    "test_unpatch_unpatch_import_patch",
                    "test_unpatch_unpatch_patch_import",
                ]
            )

            for test_attr in tests:
                if test_attr in test_ignore:
                    continue
                assert hasattr(self, test_attr), "{} not found in expected test attrs".format(test_attr)

        def assert_module_patched(self, module):
            """
            Asserts that the given module is patched.

            For example, the redis integration patches the following methods:
                - redis.StrictRedis.execute_command
                - redis.StrictRedis.pipeline
                - redis.Redis.pipeline
                - redis.client.BasePipeline.execute
                - redis.client.BasePipeline.immediate_execute_command

            So an appropriate assert_module_patched would look like::

                def assert_module_patched(self, redis):
                    self.assert_wrapped(redis.StrictRedis.execute_command)
                    self.assert_wrapped(redis.StrictRedis.pipeline)
                    self.assert_wrapped(redis.Redis.pipeline)
                    self.assert_wrapped(redis.client.BasePipeline.execute)
                    self.assert_wrapped(redis.client.BasePipeline.immediate_execute_command)

            :param module: module to check
            :return: None
            """
            raise NotImplementedError(self.assert_module_patched.__doc__)

        def assert_not_module_patched(self, module):
            """
            Asserts that the given module is not patched.

            For example, the redis integration patches the following methods:
                - redis.StrictRedis.execute_command
                - redis.StrictRedis.pipeline
                - redis.Redis.pipeline
                - redis.client.BasePipeline.execute
                - redis.client.BasePipeline.immediate_execute_command

            So an appropriate assert_not_module_patched would look like::

                def assert_not_module_patched(self, redis):
                    self.assert_not_wrapped(redis.StrictRedis.execute_command)
                    self.assert_not_wrapped(redis.StrictRedis.pipeline)
                    self.assert_not_wrapped(redis.Redis.pipeline)
                    self.assert_not_wrapped(redis.client.BasePipeline.execute)
                    self.assert_not_wrapped(redis.client.BasePipeline.immediate_execute_command)

            :param module:
            :return: None
            """
            raise NotImplementedError(self.assert_not_module_patched.__doc__)

        def assert_not_module_double_patched(self, module):
            """
            Asserts that the given module is not patched twice.

            For example, the redis integration patches the following methods:
                - redis.StrictRedis.execute_command
                - redis.StrictRedis.pipeline
                - redis.Redis.pipeline
                - redis.client.BasePipeline.execute
                - redis.client.BasePipeline.immediate_execute_command

            So an appropriate assert_not_module_double_patched would look like::

                def assert_not_module_double_patched(self, redis):
                    self.assert_not_double_wrapped(redis.StrictRedis.execute_command)
                    self.assert_not_double_wrapped(redis.StrictRedis.pipeline)
                    self.assert_not_double_wrapped(redis.Redis.pipeline)
                    self.assert_not_double_wrapped(redis.client.BasePipeline.execute)
                    self.assert_not_double_wrapped(redis.client.BasePipeline.immediate_execute_command)

            :param module: module to check
            :return: None
            """
            raise NotImplementedError(self.assert_not_module_double_patched.__doc__)

        @raise_if_no_attrs
        def test_import_patch(self):
            """
            The integration should test that each class, method or function that
            is to be patched is in fact done so when ddtrace.patch() is called
            before the module is imported.

            For example:

            an appropriate ``test_patch_import`` would be::

                import redis
                ddtrace.patch(redis=True)
                self.assert_module_patched(redis)
            """
            module = importlib.import_module(self.__module_name__)
            self.assert_not_module_patched(module)
            self.__patch_func__()
            self.assert_module_patched(module)

        @raise_if_no_attrs
        def test_patch_import(self):
            """
            The integration should test that each class, method or function that
            is to be patched is in fact done so when ddtrace.patch() is called
            after the module is imported.

            an appropriate ``test_patch_import`` would be::

                import redis
                ddtrace.patch(redis=True)
                self.assert_module_patched(redis)
            """
            module = importlib.import_module(self.__module_name__)
            self.__patch_func__()
            self.assert_module_patched(module)

        @raise_if_no_attrs
        def test_import_patch_patch(self):
            """
            Proper testing should be done to ensure that multiple calls to the
            integration.patch() method are idempotent. That is, that the
            integration does not patch its library more than once.

            An example for what this might look like for the redis integration::

                import redis
                ddtrace.patch(redis=True)
                self.assert_module_patched(redis)
                ddtrace.patch(redis=True)
                self.assert_not_module_double_patched(redis)
            """
            self.__patch_func__()
            module = importlib.import_module(self.__module_name__)
            self.assert_module_patched(module)
            self.__patch_func__()
            self.assert_not_module_double_patched(module)

        @raise_if_no_attrs
        def test_patch_import_patch(self):
            """
            Proper testing should be done to ensure that multiple calls to the
            integration.patch() method are idempotent. That is, that the
            integration does not patch its library more than once.

            An example for what this might look like for the redis integration::

                ddtrace.patch(redis=True)
                import redis
                self.assert_module_patched(redis)
                ddtrace.patch(redis=True)
                self.assert_not_module_double_patched(redis)
            """
            self.__patch_func__()
            module = importlib.import_module(self.__module_name__)
            self.assert_module_patched(module)
            self.__patch_func__()
            self.assert_not_module_double_patched(module)

        @raise_if_no_attrs
        def test_patch_patch_import(self):
            """
            Proper testing should be done to ensure that multiple calls to the
            integration.patch() method are idempotent. That is, that the
            integration does not patch its library more than once.

            An example for what this might look like for the redis integration::

                ddtrace.patch(redis=True)
                ddtrace.patch(redis=True)
                import redis
                self.assert_not_double_wrapped(redis.StrictRedis.execute_command)
            """
            self.__patch_func__()
            self.__patch_func__()
            module = importlib.import_module(self.__module_name__)
            self.assert_module_patched(module)
            self.assert_not_module_double_patched(module)

        @noop_if_no_unpatch
        @raise_if_no_attrs
        def test_import_patch_unpatch_patch(self):
            """
            To ensure that we can thoroughly test the installation/patching of
            an integration we must be able to unpatch it and then subsequently
            patch it again.

            For example::

                import redis
                from ddtrace.contrib.redis import unpatch

                ddtrace.patch(redis=True)
                unpatch()
                ddtrace.patch(redis=True)
                self.assert_module_patched(redis)
            """
            module = importlib.import_module(self.__module_name__)
            self.__patch_func__()
            self.assert_module_patched(module)
            self.__unpatch_func__()
            self.assert_not_module_patched(module)
            self.__patch_func__()
            self.__patch_func__()
            self.assert_module_patched(module)

        @noop_if_no_unpatch
        @raise_if_no_attrs
        def test_patch_import_unpatch_patch(self):
            """
            To ensure that we can thoroughly test the installation/patching of
            an integration we must be able to unpatch it and then subsequently
            patch it again.

            For example::

                from ddtrace.contrib.redis import unpatch

                ddtrace.patch(redis=True)
                import redis
                unpatch()
                ddtrace.patch(redis=True)
                self.assert_module_patched(redis)
            """
            self.__patch_func__()
            module = importlib.import_module(self.__module_name__)
            self.assert_module_patched(module)
            self.__unpatch_func__()
            self.assert_not_module_patched(module)
            self.__patch_func__()
            self.assert_module_patched(module)

        @noop_if_no_unpatch
        @raise_if_no_attrs
        def test_patch_unpatch_import_patch(self):
            """
            To ensure that we can thoroughly test the installation/patching of
            an integration we must be able to unpatch it and then subsequently
            patch it again.

            For example::

                from ddtrace.contrib.redis import unpatch

                ddtrace.patch(redis=True)
                import redis
                unpatch()
                ddtrace.patch(redis=True)
                self.assert_module_patched(redis)
            """
            self.__patch_func__()
            self.__unpatch_func__()
            module = importlib.import_module(self.__module_name__)
            self.assert_not_module_patched(module)
            self.__patch_func__()
            self.assert_module_patched(module)

        @noop_if_no_unpatch
        @raise_if_no_attrs
        def test_patch_unpatch_patch_import(self):
            """
            To ensure that we can thoroughly test the installation/patching of
            an integration we must be able to unpatch it and then subsequently
            patch it again.

            For example::

                from ddtrace.contrib.redis import unpatch

                ddtrace.patch(redis=True)
                unpatch()
                ddtrace.patch(redis=True)
                import redis
                self.assert_module_patched(redis)
            """
            self.__patch_func__()
            self.__unpatch_func__()
            self.__patch_func__()
            module = importlib.import_module(self.__module_name__)
            self.assert_module_patched(module)

        @noop_if_no_unpatch
        @raise_if_no_attrs
        def test_unpatch_patch_import(self):
            """
            Make sure unpatching before patch does not break patching.

            For example::

                from ddtrace.contrib.redis import unpatch
                unpatch()
                ddtrace.patch(redis=True)
                import redis
                self.assert_not_module_patched(redis)
            """
            self.__unpatch_func__()
            self.__patch_func__()
            self.__patch_func__()
            module = importlib.import_module(self.__module_name__)
            self.assert_module_patched(module)

        @noop_if_no_unpatch
        @raise_if_no_attrs
        def test_patch_unpatch_import(self):
            """
            To ensure that we can thoroughly test the installation/patching of
            an integration we must be able to unpatch it before importing the
            library.

            For example::

                ddtrace.patch(redis=True)
                from ddtrace.contrib.redis import unpatch
                unpatch()
                import redis
                self.assert_not_module_patched(redis)
            """
            self.__patch_func__()
            self.__unpatch_func__()
            module = importlib.import_module(self.__module_name__)
            self.assert_not_module_patched(module)

        @noop_if_no_unpatch
        @raise_if_no_attrs
        def test_import_unpatch_patch(self):
            """
            To ensure that we can thoroughly test the installation/patching of
            an integration we must be able to unpatch it before patching.

            For example::

                import redis
                from ddtrace.contrib.redis import unpatch
                ddtrace.patch(redis=True)
                unpatch()
                self.assert_not_module_patched(redis)
            """
            module = importlib.import_module(self.__module_name__)
            self.__unpatch_func__()
            self.assert_not_module_patched(module)
            self.__patch_func__()
            self.assert_module_patched(module)

        @noop_if_no_unpatch
        @raise_if_no_attrs
        def test_import_patch_unpatch(self):
            """
            To ensure that we can thoroughly test the installation/patching of
            an integration we must be able to unpatch it after patching.

            For example::

                import redis
                from ddtrace.contrib.redis import unpatch
                ddtrace.patch(redis=True)
                unpatch()
                self.assert_not_module_patched(redis)
            """
            module = importlib.import_module(self.__module_name__)
            self.assert_not_module_patched(module)
            self.__patch_func__()
            self.assert_module_patched(module)
            self.__unpatch_func__()
            self.assert_not_module_patched(module)

        @noop_if_no_unpatch
        @raise_if_no_attrs
        def test_patch_import_unpatch(self):
            """
            To ensure that we can thoroughly test the installation/patching of
            an integration we must be able to unpatch it after patching.

            For example::

                from ddtrace.contrib.redis import unpatch
                ddtrace.patch(redis=True)
                import redis
                unpatch()
                self.assert_not_module_patched(redis)
            """
            self.__patch_func__()
            module = importlib.import_module(self.__module_name__)
            self.assert_module_patched(module)
            self.__unpatch_func__()
            self.assert_not_module_patched(module)

        @noop_if_no_unpatch
        @raise_if_no_attrs
        def test_import_patch_unpatch_unpatch(self):
            """
            Unpatching twice should be a no-op.

            For example::

                import redis
                from ddtrace.contrib.redis import unpatch

                ddtrace.patch(redis=True)
                self.assert_module_patched(redis)
                unpatch()
                self.assert_not_module_patched(redis)
                unpatch()
                self.assert_not_module_patched(redis)
            """
            module = importlib.import_module(self.__module_name__)
            self.__patch_func__()
            self.assert_module_patched(module)
            self.__unpatch_func__()
            self.assert_not_module_patched(module)
            self.__unpatch_func__()
            self.assert_not_module_patched(module)

        @noop_if_no_unpatch
        @raise_if_no_attrs
        def test_patch_unpatch_import_unpatch(self):
            """
            Unpatching twice should be a no-op.

            For example::

                from ddtrace.contrib.redis import unpatch

                ddtrace.patch(redis=True)
                unpatch()
                import redis
                self.assert_not_module_patched(redis)
                unpatch()
                self.assert_not_module_patched(redis)
            """
            self.__patch_func__()
            self.__unpatch_func__()
            module = importlib.import_module(self.__module_name__)
            self.assert_not_module_patched(module)
            self.__unpatch_func__()
            self.assert_not_module_patched(module)

        @noop_if_no_unpatch
        @raise_if_no_attrs
        def test_patch_unpatch_unpatch_import(self):
            """
            Unpatching twice should be a no-op.

            For example::

                from ddtrace.contrib.redis import unpatch

                ddtrace.patch(redis=True)
                unpatch()
                unpatch()
                import redis
                self.assert_not_module_patched(redis)
            """
            self.__patch_func__()
            self.__unpatch_func__()
            self.__unpatch_func__()
            module = importlib.import_module(self.__module_name__)
            self.assert_not_module_patched(module)



"""
The Sanic__ integration will trace requests to and from Sanic.


Enable Sanic tracing automatically via ``ddtrace-run``::

    ddtrace-run python app.py

Sanic tracing can also be enabled manually::

    from ddtrace import patch_all
    patch_all(sanic=True)

    from sanic import Sanic
    from sanic.response import text

    app = Sanic(__name__)

    @app.route('/')
    def index(request):
        return text('hello world')

    if __name__ == '__main__':
        app.run()

If using Python 3.6, the legacy ``AsyncioContextProvider`` will have to be
enabled before using the middleware::

    from ddtrace.contrib.asyncio.provider import AsyncioContextProvider
    from ddtrace import tracer  # Or whichever tracer instance you plan to use
    tracer.configure(context_provider=AsyncioContextProvider())


Configuration
~~~~~~~~~~~~~

.. py:data:: ddtrace.config.sanic['distributed_tracing_enabled']

   Whether to parse distributed tracing headers from requests received by your Sanic app.

   Default: ``True``


.. py:data:: ddtrace.config.sanic['service_name']

   The service name reported for your Sanic app.

   Can also be configured via the ``DD_SERVICE`` environment variable.

   Default: ``'sanic'``


Example::

    from ddtrace import config

    # Enable distributed tracing
    config.sanic['distributed_tracing_enabled'] = True

    # Override service name
    config.sanic['service_name'] = 'custom-service-name'

.. __: https://sanic.readthedocs.io/en/latest/
"""
from ...internal.utils.importlib import require_modules


required_modules = ["sanic"]

with require_modules(required_modules) as missing_modules:
    if not missing_modules:
        from .patch import patch
        from .patch import unpatch

        __all__ = ["patch", "unpatch"]



# -*- coding: utf-8 -*-

#    Copyright (C) 2015 Yahoo! Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from .. import wrapt

from . import _utils

_KWARG_RENAMED_POSTFIX_TPL = ", please use the '%s' argument instead"
_KWARG_RENAMED_PREFIX_TPL = "Using the '%s' argument is deprecated"


def renamed_kwarg(old_name, new_name, message=None,
                  version=None, removal_version=None, stacklevel=3,
                  category=None, replace=False):
    """Decorates a kwarg accepting function to deprecate a renamed kwarg."""

    prefix = _KWARG_RENAMED_PREFIX_TPL % old_name
    postfix = _KWARG_RENAMED_POSTFIX_TPL % new_name
    out_message = _utils.generate_message(
        prefix, postfix=postfix, message=message, version=version,
        removal_version=removal_version)

    @wrapt.decorator
    def decorator(wrapped, instance, args, kwargs):
        if old_name in kwargs:
            _utils.deprecation(out_message,
                               stacklevel=stacklevel, category=category)
            if replace:
                kwargs.setdefault(new_name, kwargs.pop(old_name))
        return wrapped(*args, **kwargs)

    return decorator



# Unless explicitly stated otherwise all files in this repository are licensed under the BSD-3-Clause License.
# This product includes software developed at Datadog (https://www.datadoghq.com/).
# Copyright 2015-Present Datadog, Inc
from .base import DogStatsd, statsd  # noqa



from types import CodeType
from types import FunctionType
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import cast

from ddtrace.debugging._function.discovery import FullyNamed
from ddtrace.internal.injection import HookInfoType
from ddtrace.internal.injection import HookType
from ddtrace.internal.injection import eject_hooks
from ddtrace.internal.injection import inject_hooks
from ddtrace.internal.wrapping import WrappedFunction
from ddtrace.internal.wrapping import Wrapper
from ddtrace.internal.wrapping import unwrap
from ddtrace.internal.wrapping import wrap


WrapperType = Callable[[FunctionType, Any, Any, Any], Any]


class FullyNamedWrappedFunction(FullyNamed, WrappedFunction):
    """A fully named wrapper function."""


class FunctionStore(object):
    """Function object store.

    This class provides a storage layer for patching operations, which allows us
    to store the original code object of functions being patched with either
    hook injections or wrapping. This also enforce a single wrapping layer.
    Multiple wrapping is implemented as a list of wrappers handled by the single
    wrapper function.

    If extra attributes are defined during the patching process, they will get
    removed when the functions are restored.
    """

    def __init__(self, extra_attrs=None):
        # type: (Optional[List[str]]) -> None
        self._code_map = {}  # type: Dict[FunctionType, CodeType]
        self._wrapper_map = {}  # type: Dict[FunctionType, Wrapper]
        self._extra_attrs = ["__dd_wrapped__"]
        if extra_attrs:
            self._extra_attrs.extend(extra_attrs)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.restore_all()

    def _store(self, function):
        # type: (FunctionType) -> None
        if function not in self._code_map:
            self._code_map[function] = function.__code__

    def inject_hooks(self, function, hooks):
        # type: (FullyNamedWrappedFunction, List[HookInfoType]) -> Set[str]
        """Bulk-inject hooks into a function.

        Returns the set of probe IDs for those probes that failed to inject.
        """
        try:
            return self.inject_hooks(cast(FullyNamedWrappedFunction, function.__dd_wrapped__), hooks)
        except AttributeError:
            f = cast(FunctionType, function)
            self._store(f)
            return {p.probe_id for _, _, p in inject_hooks(f, hooks)}

    def eject_hooks(self, function, hooks):
        # type: (FunctionType, List[HookInfoType]) -> Set[str]
        """Bulk-eject hooks from a function.

        Returns the set of probe IDs for those probes that failed to eject.
        """
        try:
            wrapped = cast(FullyNamedWrappedFunction, function).__dd_wrapped__
        except AttributeError:
            # Not a wrapped function so we can actually eject from it
            return {p.probe_id for _, _, p in eject_hooks(function, hooks)}
        else:
            # Try on the wrapped function.
            return self.eject_hooks(cast(FunctionType, wrapped), hooks)

    def inject_hook(self, function, hook, line, arg):
        # type: (FullyNamedWrappedFunction, HookType, int, Any) -> bool
        """Inject a hook into a function."""
        return not not self.inject_hooks(function, [(hook, line, arg)])

    def eject_hook(self, function, hook, line, arg):
        # type: (FunctionType, HookType, int, Any) -> bool
        """Eject a hook from a function."""
        return not not self.eject_hooks(function, [(hook, line, arg)])

    def wrap(self, function, wrapper):
        # type: (FunctionType, Wrapper) -> None
        """Wrap a function with a hook."""
        self._store(function)
        self._wrapper_map[function] = wrapper
        wrap(function, wrapper)

    def unwrap(self, function):
        # type: (FullyNamedWrappedFunction) -> None
        """Unwrap a hook around a wrapped function."""
        unwrap(function, self._wrapper_map.pop(cast(FunctionType, function)))

    def restore_all(self):
        # type: () -> None
        """Restore all the patched functions to their original form."""
        for function, code in self._code_map.items():
            function.__code__ = code
            for attr in self._extra_attrs:
                try:
                    delattr(function, attr)
                except AttributeError:
                    pass



from itertools import chain
import time
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Optional

from ddtrace import config as tracer_config
from ddtrace.debugging._config import config
from ddtrace.debugging._expressions import dd_compile
from ddtrace.debugging._probe.model import FunctionProbe
from ddtrace.debugging._probe.model import LineProbe
from ddtrace.debugging._probe.model import MetricProbe
from ddtrace.debugging._probe.model import Probe
from ddtrace.internal.logger import get_logger
from ddtrace.internal.utils.cache import LFUCache


log = get_logger(__name__)


_EXPRESSION_CACHE = LFUCache()


def _invalid_condition(_):
    """Forces probes with invalid conditions to never trigger.

    Any signs of invalid conditions in logs is an indication of a problem with
    the expression compiler.
    """
    return False


INVALID_CONDITION = _invalid_condition


def _compile_condition(when):
    # type: (Optional[Dict[str, Any]]) -> Optional[Callable[[Dict[str, Any]], Any]]
    global _EXPRESSION_CACHE, INVALID_CONDITION

    if when is None:
        return None

    ast = when["json"]

    def compile_or_invalid(expr):
        # type: (str) -> Callable[[Dict[str, Any]], Any]
        try:
            return dd_compile(ast)
        except Exception:
            log.error("Cannot compile expression: %s", expr, exc_info=True)
            return INVALID_CONDITION

    expr = when["dsl"]

    compiled = _EXPRESSION_CACHE.get(expr, compile_or_invalid)  # type: Callable[[Dict[str, Any]], Any]

    if compiled is INVALID_CONDITION:
        log.error("Cannot compile expression: %s", expr, exc_info=True)

    return compiled


def _match_env_and_version(probe):
    # type: (Probe) -> bool
    probe_version = probe.tags.get("version", None)
    probe_env = probe.tags.get("env", None)

    return (probe_version is None or probe_version == tracer_config.version) and (
        probe_env is None or probe_env == tracer_config.env
    )


def _filter_by_env_and_version(f):
    # type: (Callable[..., Iterable[Probe]]) -> Callable[..., Iterable[Probe]]
    def _wrapper(*args, **kwargs):
        # type: (Any, Any) -> Iterable[Probe]
        return [_ for _ in f(*args, **kwargs) if _match_env_and_version(_)]

    return _wrapper


def probe(_id, _type, attribs):
    # type: (str, str, Dict[str, Any]) -> Probe
    """
    Create a new Probe instance.
    """
    if _type == "snapshotProbes":
        args = dict(
            probe_id=_id,
            condition=_compile_condition(attribs.get("when")),
            active=attribs["active"],
            tags=dict(_.split(":", maxsplit=1) for _ in attribs.get("tags", [])),
        )

        if attribs["where"].get("sourceFile", None):
            ProbeType = LineProbe
            args["source_file"] = attribs["where"]["sourceFile"]
            args["line"] = int(attribs["where"]["lines"][0])
        else:
            ProbeType = FunctionProbe  # type: ignore[assignment]
            args["module"] = attribs["where"].get("type") or attribs["where"]["typeName"]
            args["func_qname"] = attribs["where"].get("method") or attribs["where"]["methodName"]

        return ProbeType(**args)

    elif _type == "metricProbes":
        return MetricProbe(
            probe_id=_id,
            active=attribs["active"],
            tags=dict(_.split(":", maxsplit=1) for _ in attribs.get("tags", [])),
            source_file=attribs["where"]["sourceFile"],
            line=int(attribs["where"]["lines"][0]),
            name=attribs["metricName"],
            kind=attribs["kind"],
        )

    raise ValueError("Unknown probe type: %s" % _type)


@_filter_by_env_and_version
def get_probes(service_config):
    # type: (dict) -> Iterable[Probe]
    return chain(
        [probe(p["id"], "snapshotProbes", p) for p in service_config.get("snapshotProbes") or []],
        [probe(p["id"], "metricProbes", p) for p in service_config.get("metricProbes") or []],
    )


log = get_logger(__name__)


class ProbePollerEvent(object):
    NEW_PROBES = 0
    DELETED_PROBES = 1
    MODIFIED_PROBES = 2
    STATUS_UPDATE = 3


ProbePollerEventType = int


def probe_modified(reference, probe):
    # type: (Probe, Probe) -> bool
    # DEV: Probes are immutable modulo the active state. Hence we expect
    # probes with the same ID to differ up to this property for now.
    return probe.active != reference.active


class ProbeRCAdapter(object):
    """Probe configuration adapter for the RCM client.

    This adapter turns configuration events from the RCM client into probe
    events that can be handled easily by the debugger.
    """

    def __init__(self, callback):
        # type: (Callable[[ProbePollerEventType, Iterable[Probe]], None]) -> None
        self._callback = callback
        self._probes = {}  # type: Dict[str, Probe]
        self._next_status_update_timestamp()

    def _next_status_update_timestamp(self):
        # type: () -> None
        self._status_timestamp = time.time() + config.diagnostics_interval

    def __call__(self, metadata, config):
        # type: (Any, Any) -> None
        # DEV: We emit a status update event here to avoid having to spawn a
        # separate thread for this.
        if time.time() > self._status_timestamp:
            log.debug("Emitting probe status log messages")
            self._callback(ProbePollerEvent.STATUS_UPDATE, self._probes.values())
            self._next_status_update_timestamp()

        if config is None:
            return

        probes = get_probes(config)

        current_probes = {_.probe_id: _ for _ in probes}

        new_probes = [p for _, p in current_probes.items() if _ not in self._probes]
        deleted_probes = [p for _, p in self._probes.items() if _ not in current_probes]
        modified_probes = [
            p for _, p in current_probes.items() if _ in self._probes and probe_modified(p, self._probes[_])
        ]

        if deleted_probes:
            self._callback(ProbePollerEvent.DELETED_PROBES, deleted_probes)
        if modified_probes:
            self._callback(ProbePollerEvent.MODIFIED_PROBES, modified_probes)
        if new_probes:
            self._callback(ProbePollerEvent.NEW_PROBES, new_probes)

        self._probes = current_probes



import fastapi
from fastapi.middleware import Middleware
import fastapi.routing

from ddtrace import Pin
from ddtrace import config
from ddtrace.contrib.asgi.middleware import TraceMiddleware
from ddtrace.contrib.starlette.patch import get_resource
from ddtrace.contrib.starlette.patch import traced_handler
from ddtrace.internal.logger import get_logger
from ddtrace.internal.utils.deprecations import DDTraceDeprecationWarning
from ddtrace.internal.utils.wrappers import unwrap as _u
from ddtrace.vendor.debtcollector import removals
from ddtrace.vendor.wrapt import ObjectProxy
from ddtrace.vendor.wrapt import wrap_function_wrapper as _w


log = get_logger(__name__)

config._add(
    "fastapi",
    dict(
        _default_service="fastapi",
        request_span_name="fastapi.request",
        distributed_tracing=True,
        aggregate_resources=True,
    ),
)


@removals.remove(removal_version="2.0.0", category=DDTraceDeprecationWarning)
def span_modifier(span, scope):
    resource = get_resource(scope)
    if config.fastapi["aggregate_resources"] and resource:
        span.resource = "{} {}".format(scope["method"], resource)


def traced_init(wrapped, instance, args, kwargs):
    mw = kwargs.pop("middleware", [])
    mw.insert(0, Middleware(TraceMiddleware, integration_config=config.fastapi))
    kwargs.update({"middleware": mw})
    wrapped(*args, **kwargs)


async def traced_serialize_response(wrapped, instance, args, kwargs):
    """Wrapper for fastapi.routing.serialize_response function.

    This function is called on all non-Response objects to
    convert them to a serializable form.

    This is the wrapper which calls ``jsonable_encoder``.

    This function does not do the actual encoding from
    obj -> json string  (e.g. json.dumps()). That is handled
    by the Response.render function.

    DEV: We do not wrap ``jsonable_encoder`` because it calls
    itself recursively, so there is a chance the overhead
    added by creating spans will be higher than desired for
    the result.
    """
    pin = Pin.get_from(fastapi)
    if not pin or not pin.enabled:
        return await wrapped(*args, **kwargs)

    with pin.tracer.trace("fastapi.serialize_response"):
        return await wrapped(*args, **kwargs)


def patch():
    if getattr(fastapi, "_datadog_patch", False):
        return

    setattr(fastapi, "_datadog_patch", True)
    Pin().onto(fastapi)
    _w("fastapi.applications", "FastAPI.__init__", traced_init)
    _w("fastapi.routing", "serialize_response", traced_serialize_response)

    # We need to check that Starlette instrumentation hasn't already patched these
    if not isinstance(fastapi.routing.APIRoute.handle, ObjectProxy):
        _w("fastapi.routing", "APIRoute.handle", traced_handler)

    if not isinstance(fastapi.routing.Mount.handle, ObjectProxy):
        _w("starlette.routing", "Mount.handle", traced_handler)


def unpatch():
    if not getattr(fastapi, "_datadog_patch", False):
        return

    setattr(fastapi, "_datadog_patch", False)

    _u(fastapi.applications.FastAPI, "__init__")
    _u(fastapi.routing, "serialize_response")

    # We need to check that Starlette instrumentation hasn't already unpatched these
    if isinstance(fastapi.routing.APIRoute.handle, ObjectProxy):
        _u(fastapi.routing.APIRoute, "handle")

    if isinstance(fastapi.routing.Mount.handle, ObjectProxy):
        _u(fastapi.routing.Mount, "handle")



# -*- coding: utf-8 -*-
import os.path

# 3rd party
import jinja2

from ddtrace import Pin
from ddtrace import config
from ddtrace.contrib.jinja2 import patch
from ddtrace.contrib.jinja2 import unpatch
from tests.utils import TracerTestCase
from tests.utils import assert_is_measured
from tests.utils import assert_is_not_measured


TEST_DIR = os.path.dirname(os.path.realpath(__file__))
TMPL_DIR = os.path.join(TEST_DIR, "templates")


class Jinja2Test(TracerTestCase):
    def setUp(self):
        super(Jinja2Test, self).setUp()
        patch()
        # prevent cache effects when using Template('code...')
        try:
            jinja2.environment._spontaneous_environments.clear()
        except AttributeError:
            jinja2.utils.clear_caches()
        Pin.override(jinja2.environment.Environment, tracer=self.tracer)

    def tearDown(self):
        super(Jinja2Test, self).tearDown()
        # restore the tracer
        unpatch()

    def test_render_inline_template(self):
        t = jinja2.environment.Template("Hello {{name}}!")
        assert t.render(name="Jinja") == "Hello Jinja!"

        # tests
        spans = self.pop_spans()
        assert len(spans) == 2

        for span in spans:
            assert span.service is None
            assert span.span_type == "template"
            assert span.get_tag("jinja2.template_name") == "<memory>"

        assert spans[0].name == "jinja2.compile"
        assert_is_not_measured(spans[0])
        assert spans[1].name == "jinja2.render"
        assert_is_measured(spans[1])

    def test_generate_inline_template(self):
        t = jinja2.environment.Template("Hello {{name}}!")
        assert "".join(t.generate(name="Jinja")) == "Hello Jinja!"

        # tests
        spans = self.pop_spans()
        assert len(spans) == 2

        for span in spans:
            assert span.service is None
            assert span.span_type == "template"
            assert span.get_tag("jinja2.template_name") == "<memory>"

        assert spans[0].name == "jinja2.compile"
        assert_is_not_measured(spans[0])
        assert spans[1].name == "jinja2.render"
        assert_is_measured(spans[1])

    def test_custom_template_name(self):
        """
        Regression test for #4008

        https://github.com/DataDog/dd-trace-py/issues/4008
        """
        # pytest.mark.parametrize is not supported for unittest test methods
        cases = [
            ("custom-name", "custom-name"),
            (1, "1"),
            (u"😐", u"😐"),
        ]
        for template_name, expected in cases:
            t = jinja2.environment.Template("Hello {{name}}!")
            t.name = template_name
            assert "".join(t.generate(name="Jinja")) == "Hello Jinja!"

            # tests
            spans = self.pop_spans()
            assert len(spans) == 2

            render_span = spans[1]
            assert render_span.name == "jinja2.render"
            assert render_span.get_tag("jinja2.template_name") == expected
            assert render_span.resource == expected

    def test_file_template(self):
        loader = jinja2.loaders.FileSystemLoader(TMPL_DIR)
        env = jinja2.Environment(loader=loader)
        t = env.get_template("template.html")
        assert t.render(name="Jinja") == "Message: Hello Jinja!"

        # tests
        spans = self.pop_spans()
        assert len(spans) == 5

        for span in spans:
            assert span.span_type == "template"
            assert span.service is None

        # templates.html extends base.html
        def get_def(s):
            return s.name, s.get_tag("jinja2.template_name")

        assert get_def(spans[0]) == ("jinja2.load", "template.html")
        assert_is_not_measured(spans[0])
        assert get_def(spans[1]) == ("jinja2.compile", "template.html")
        assert_is_not_measured(spans[1])
        assert get_def(spans[2]) == ("jinja2.render", "template.html")
        assert_is_measured(spans[2])
        assert get_def(spans[3]) == ("jinja2.load", "base.html")
        assert_is_not_measured(spans[3])
        assert get_def(spans[4]) == ("jinja2.compile", "base.html")
        assert_is_not_measured(spans[4])

        # additional checks for jinja2.load
        assert spans[0].get_tag("jinja2.template_path") == os.path.join(TMPL_DIR, "template.html")
        assert spans[3].get_tag("jinja2.template_path") == os.path.join(TMPL_DIR, "base.html")

    def test_service_name(self):
        # don't inherit the service name from the parent span, but force the value.
        loader = jinja2.loaders.FileSystemLoader(TMPL_DIR)
        env = jinja2.Environment(loader=loader)

        cfg = config.get_from(env)
        cfg["service_name"] = "renderer"

        t = env.get_template("template.html")
        assert t.render(name="Jinja") == "Message: Hello Jinja!"

        # tests
        spans = self.pop_spans()
        assert len(spans) == 5

        for span in spans:
            assert span.service == "renderer"

    def test_inherit_service(self):
        # When there is a parent span and no custom service_name, the service name is inherited
        loader = jinja2.loaders.FileSystemLoader(TMPL_DIR)
        env = jinja2.Environment(loader=loader)

        with self.tracer.trace("parent.span", service="web"):
            t = env.get_template("template.html")
            assert t.render(name="Jinja") == "Message: Hello Jinja!"

        # tests
        spans = self.pop_spans()
        assert len(spans) == 6

        for span in spans:
            assert span.service == "web"

    @TracerTestCase.run_in_subprocess(env_overrides=dict(DD_SERVICE="mysvc"))
    def test_user_specified_service(self):
        """
        When a service name is specified by the user
            The jinja integration should use it as the service name
        """

        loader = jinja2.loaders.FileSystemLoader(TMPL_DIR)
        env = jinja2.Environment(loader=loader)
        t = env.get_template("template.html")
        assert t.render(name="Jinja") == "Message: Hello Jinja!"

        spans = self.pop_spans()
        for span in spans:
            assert span.service == "mysvc"
